from policyengine_us.model_api import *


class income_adjusted_part_b_premium(Variable):
    value_type = float
    entity = Person
    label = "Medicare Part B premium (income-adjusted)"
    unit = USD
    definition_period = YEAR
    defined_for = "is_medicare_eligible"
    reference = "https://www.medicare.gov/your-medicare-costs/part-b-costs"
    documentation = "Medicare Part B premium adjusted for income (IRMAA). Based on modified adjusted gross income from 2 years prior."

    def formula(person, period, parameters):

        tax_unit = person.tax_unit
        filing_status = tax_unit("filing_status", period)
        # Medicare Part B IRMAA is based on MAGI from 2 years prior
        # MAGI = AGI + tax-exempt interest
        prior_period = period.offset(-2, "year")
        agi = tax_unit("adjusted_gross_income", prior_period)
        tax_exempt_interest = tax_unit(
            "tax_exempt_interest_income", prior_period
        )
        magi = agi + tax_exempt_interest
        base = person("base_part_b_premium", period)

        # Build boolean masks for each status
        status = filing_status.possible_values
        statuses = [
            status.SINGLE,
            status.JOINT,
            status.HEAD_OF_HOUSEHOLD,
            status.SURVIVING_SPOUSE,
            status.SEPARATE,
        ]
        in_status = [filing_status == s for s in statuses]

        p = parameters(period).gov.hhs.medicare.part_b.irmaa

        irmaa_amount = select(
            in_status,
            [
                p.single.calc(magi),
                p.joint.calc(magi),
                p.head_of_household.calc(magi),
                p.surviving_spouse.calc(magi),
                p.separate.calc(magi),
            ],
        )

        # IRMAA amounts are monthly, multiply by MONTHS_IN_YEAR to get annual
        # Base is already annual
        annual_irmaa = irmaa_amount * MONTHS_IN_YEAR
        return base + annual_irmaa



from policyengine_us.model_api import *


class nyc_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC income tax before credits"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("nyc_taxable_income", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        rates = parameters(period).gov.local.ny.nyc.tax.income.rates
        return select(
            [
                filing_status == filing_statuses.SINGLE,
                filing_status == filing_statuses.JOINT,
                filing_status == filing_statuses.HEAD_OF_HOUSEHOLD,
                filing_status == filing_statuses.SURVIVING_SPOUSE,
                filing_status == filing_statuses.SEPARATE,
            ],
            [
                rates.single.calc(taxable_income),
                rates.joint.calc(taxable_income),
                rates.head_of_household.calc(taxable_income),
                rates.surviving_spouse.calc(taxable_income),
                rates.separate.calc(taxable_income),
            ],
        )



from policyengine_us.model_api import *


def create_or_rebate_state_tax_exempt() -> Reform:
    class or_rebate_subtraction(Variable):
        value_type = float
        entity = TaxUnit
        label = "Oregon rebate subtraction"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.OR

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.ubi_center.basic_income
            if p.taxable:
                return tax_unit("basic_income", period)
            return 0

    class or_income_subtractions(Variable):
        value_type = float
        entity = TaxUnit
        label = "OR income subtractions"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=13",
            "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # Subsection 316.800
        )
        defined_for = StateCode.OR

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states["or"].tax.income.subtractions
            previous_subtractions = add(tax_unit, period, p.subtractions)
            rebate_subtraction = tax_unit("or_rebate_subtraction", period)
            return previous_subtractions + rebate_subtraction

    class reform(Reform):
        def apply(self):
            self.update_variable(or_income_subtractions)
            self.update_variable(or_rebate_subtraction)

    return reform


def create_or_rebate_state_tax_exempt_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_or_rebate_state_tax_exempt()

    p = parameters(period).gov.contrib.states["or"].rebate

    if p.state_tax_exempt:
        return create_or_rebate_state_tax_exempt()
    else:
        return None


or_rebate_state_tax_exempt = create_or_rebate_state_tax_exempt_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class ia_taxable_income_modifications_consolidated(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa modifications to taxable income for years on or after 2023"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.iowa.gov/media/2746/download?inline",
        "https://www.legis.iowa.gov/docs/code/422.7.pdf",
    )
    defined_for = StateCode.IA

    adds = ["ia_additions_consolidated"]
    subtracts = ["ia_subtractions_consolidated"]



from policyengine_us.model_api import *


class dc_ccsp_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC Child Care Subsidy Program (CCSP) due to income"
    definition_period = MONTH
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=12"
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp.income.income_limit
        countable_income = spm_unit("dc_ccsp_countable_income", period)
        fpg = spm_unit("spm_unit_fpg", period)
        smi = spm_unit("hhs_smi", period)
        enrolled_in_ccsp = spm_unit("dc_ccsp_enrolled", period)
        new_applicants_income_limit = fpg * p.new_applicants_rate
        redetermination_income_limit = smi * p.redetermination_rate
        income_limit = where(
            enrolled_in_ccsp,
            redetermination_income_limit,
            new_applicants_income_limit,
        )
        return countable_income <= income_limit



from policyengine_us.model_api import *


class ma_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA income tax"
    unit = USD
    documentation = "Massachusetts State income tax."
    definition_period = YEAR
    reference = "https://www.mass.gov/doc/2021-form-1-massachusetts-resident-income-tax-return/download"
    defined_for = StateCode.MA
    adds = ["ma_income_tax_before_refundable_credits"]
    subtracts = ["ma_refundable_credits"]



from policyengine_us.model_api import *


class sc_military_deduction_indv(Variable):
    value_type = float
    entity = Person
    label = "South Carolina military deduction for eligible individuals"
    unit = USD
    reference = (
        "https://www.scstatehouse.gov/code/t12c006.php",  # SECTION 12-6-1171(A)
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2021.pdf#page=17",
    )
    definition_period = YEAR
    defined_for = StateCode.SC

    def formula(person, period, parameters):
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return person("military_retirement_pay", period) * head_or_spouse



from policyengine_us.model_api import *


class tax_unit_taxable_unemployment_compensation(Variable):
    value_type = float
    entity = TaxUnit
    label = "Taxable unemployment compensation"
    unit = USD
    documentation = "Unemployment compensation included in AGI."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/85"

    def formula(tax_unit, period, parameters):
        ui = parameters(period).gov.irs.unemployment_compensation
        uc_amount = tax_unit("tax_unit_unemployment_compensation", period)
        agi = tax_unit("taxable_uc_agi", period)
        agi_over_uc = agi - uc_amount
        filing_status = tax_unit("filing_status", period)
        uc_excluded = where(
            agi_over_uc <= ui.exemption.cutoff[filing_status],
            min_(uc_amount, ui.exemption.amount),
            0,
        )
        return uc_amount - uc_excluded



from policyengine_us.model_api import *


class dc_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://code.dccouncil.gov/us/dc/council/code/sections/47-1806.04"  # (f)
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        return where(
            tax_unit("eitc_child_count", period) > 0,
            tax_unit("dc_eitc_with_qualifying_child", period),
            tax_unit("dc_eitc_without_qualifying_child", period),
        )



from policyengine_us.model_api import *


class ssi_claim_is_joint(Variable):
    value_type = bool
    entity = Person
    label = "SSI claim is joint"
    definition_period = YEAR

    def formula(person, period, parameters):
        abd_person = person("is_ssi_aged_blind_disabled", period)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        eligible_person = abd_person & is_head_or_spouse

        return person.marital_unit.sum(eligible_person) > 1



from policyengine_us.model_api import *


class capital_gains_28_percent_rate_gain(Variable):
    value_type = float
    entity = TaxUnit
    label = "28-percent rate gain"
    unit = USD
    documentation = "Includes collectibles and certain small business stock gains. These are taxed at a higher (28-percent) rate than other capital gains, while a proportion is excluded from taxable income."
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1(h)(4)",
        href="https://www.law.cornell.edu/uscode/text/26/1#h_4",
    )

    adds = [
        "long_term_capital_gains_on_collectibles",
        "long_term_capital_gains_on_small_business_stock",
    ]



from policyengine_us.model_api import *


class mi_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan exemptions"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",
        "https://www.legislature.mi.gov/Publications/TaxpayerGuide.pdf",
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mi.tax.income.exemptions

        # Personal Exemptions & Stillborn Exemptions
        personal_exemption = tax_unit("mi_personal_exemptions", period)

        # Disabled exemptions
        disabled_people = add(
            tax_unit, period, ["mi_disabled_exemption_eligible_person"]
        )
        disabled_exemption = disabled_people * p.disabled.amount.base

        # Disabled veteran exemptions
        disabled_veterans = add(
            tax_unit, period, ["is_fully_disabled_service_connected_veteran"]
        )
        disabled_veteran_exemption = (
            disabled_veterans * p.disabled.amount.veteran
        )

        # Dependent-on-other-return exemptions
        filing_status = tax_unit("filing_status", period)
        dependent_elsewhere = tax_unit("head_is_dependent_elsewhere", period)
        dependent_exemption = (
            dependent_elsewhere * p.dependent_on_other_return[filing_status]
        )

        # Total exemptions
        return (
            where(dependent_elsewhere, dependent_exemption, personal_exemption)
            + disabled_exemption
            + disabled_veteran_exemption
        )



from policyengine_us.model_api import *


class wi_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin Adjusted Gross Income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf"
    )
    defined_for = StateCode.WI
    adds = ["adjusted_gross_income", "wi_additions"]
    subtracts = ["wi_income_subtractions"]



from policyengine_us.model_api import *


class ma_tafdc_infant_benefit(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) infant benefit"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-705-600"
    )
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        eligible_infant = person("ma_tafdc_eligible_infant", period)
        p = parameters(period).gov.states.ma.dta.tcap.tafdc
        return eligible_infant * p.infant_amount



from policyengine_us.model_api import *


class receives_or_needs_protective_services(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Child receiving or needs protective services"



from policyengine_us.model_api import *


class tax_unit_state(Variable):
    value_type = str
    entity = TaxUnit
    label = "Tax unit State"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        return tax_unit.household("state_code_str", period)



from policyengine_us.model_api import *


class capped_home_energy_audit_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Capped home energy audit credit"
    definition_period = YEAR
    documentation = "Capped home energy audit credit"
    unit = USD
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=366"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.irs.credits.energy_efficient_home_improvement
        expenditure = tax_unit("home_energy_audit_expenditures", period)
        rate = p.rates.home_energy_audit
        uncapped = expenditure * rate
        # Cap at either the fan cap (pre-IRA) or total property cap (post-IRA).
        cap = p.cap.annual.home_energy_audit
        return min_(uncapped, cap)



from policyengine_us.model_api import *


class co_chp_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Colorado Child Health Plan Plus eligibility"
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        medicaid_eligible = person("is_medicaid_eligible", period)
        income_level = person("medicaid_income_level", period)
        p = parameters(period).gov.states.co.hcpf.chp
        in_income_range = income_level <= p.income_limit
        age = person("age", period)
        is_child = p.child.calc(age)
        is_pregnant = person("is_pregnant", period)
        is_age_eligible = is_pregnant | is_child
        return ~medicaid_eligible & is_age_eligible & in_income_range



from policyengine_us.model_api import *


class il_tanf_payment_eligible_child(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible child for Illinois Temporary Assistance for Needy Families (TANF) payment"
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.300"
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        eligible_child = person("il_tanf_eligible_child", period)
        eligible_requirements = person(
            "il_tanf_payment_eligible_requirements", period
        )
        return eligible_child & eligible_requirements



from policyengine_us.model_api import *


class ma_ccfa_maximum_reimbursement(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts Child Care Financial Assistance (CCFA) maximum reimbursement amount per child"
    unit = USD
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"
    definition_period = MONTH
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        daily_payment = person("ma_ccfa_uncapped_daily_payment", period)
        reimbursement_multiplier = person(
            "ma_ccfa_reimbursement_ratio", period
        )
        attending_days_per_month = person(
            "childcare_attending_days_per_month", period.this_year
        )
        return (
            daily_payment * reimbursement_multiplier * attending_days_per_month
        )



from policyengine_us.model_api import *


class ar_income_tax_before_non_refundable_credits_joint(Variable):
    value_type = float
    entity = Person
    label = "Arkansas income tax before non refundable credits when married filing jointly"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_AR1000F_and_AR1000NR_Instructions.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2023_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
    )
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ar.tax.income.rates.main
        taxable_income = person("ar_taxable_income_joint", period)

        main_rate = p.rate.calc(taxable_income)
        pre_reduction_tax = main_rate * taxable_income
        reduction = p.reduction.calc(taxable_income)
        total_main_rate = max_(pre_reduction_tax - reduction, 0)
        low_income_tax = person("ar_low_income_tax_joint", period)
        return min_(total_main_rate, low_income_tax)



from policyengine_us.model_api import *


class nm_other_deductions_and_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico other income deductions and exemptions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NM

    adds = "gov.states.nm.tax.income.other_deductions_and_exemptions"



from policyengine_us.model_api import *


class ca_tanf_maximum_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "California CalWORKs Maximum Aid Payment"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "https://hhsaprogramguides.sandiegocounty.gov/CalWORKS/44-300/CalWORKs_Payment_Standards/G_CalWORKs_Payment_Standards.pdf"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.ca.cdss.tanf.cash.monthly_payment
        unit_size = spm_unit("spm_unit_size", period)
        au_size = min_(unit_size, p.max_au_size)
        region1 = spm_unit.household("ca_tanf_region1", period)
        exempt = spm_unit("ca_tanf_exempt", period)
        # Indexing on exempt_name works, but not a similar string for region.
        # Use a where instead.
        exempt_name = where(exempt, "exempt", "non_exempt")
        return (
            where(
                region1,
                p.region1[exempt_name][au_size],
                p.region2[exempt_name][au_size],
            )
            * MONTHS_IN_YEAR
        )



from policyengine_us.model_api import *


class is_child(Variable):
    value_type = bool
    entity = Person
    label = "Is a child"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("age", period) < 18



from policyengine_us.model_api import *


class MassachusettsCCFAScheduleType(Enum):
    FULL_DAY = "Full Day"
    BEFORE_ONLY = "Before Only"
    AFTER_ONLY = "After Only"
    BEFORE_AND_AFTER = "Before and After"


class ma_ccfa_schedule_type(Variable):
    value_type = Enum
    entity = Person
    possible_values = MassachusettsCCFAScheduleType
    default_value = MassachusettsCCFAScheduleType.FULL_DAY
    definition_period = MONTH
    defined_for = StateCode.MA
    label = (
        "Massachusetts Child Care Financial Assistance (CCFA) schedule type"
    )
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"



from policyengine_us.model_api import *


class pr_exemptions_is_eligible_nonstudent(Variable):
    value_type = bool
    entity = Person
    label = "Puerto Rico dependent exemption eligible nonstudent"
    reference = "https://hacienda.pr.gov/sites/default/files/inst_individuals_2023.pdf#page=28"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PR

    def formula(person, period, parameters):
        # if dependent is a student, can earn gross income up to 7500 to be eligible for exemption
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.exemptions.dependent.income_limit
        is_dependent = person("pr_is_tax_unit_dependent", period)
        gross_income = person("pr_gross_income_person", period)
        income_eligibility = gross_income < p.nonstudent
        return is_dependent & income_eligibility



from policyengine_us.model_api import *
from policyengine_us.variables.gov.hhs.medicaid.costs.medicaid_group import (
    MedicaidGroup,
)


class medicaid_cost_if_enrolled(Variable):
    value_type = float
    entity = Person
    label = "Per capita Medicaid cost by eligibility group & state"
    unit = USD
    definition_period = YEAR
    defined_for = "is_medicaid_eligible"

    def formula(person, period, parameters):
        state = person.household("state_code", period)
        group = person("medicaid_group", period)

        p = parameters(period).calibration.gov.hhs.medicaid

        # Handle each group type separately
        is_aged_disabled = group == MedicaidGroup.AGED_DISABLED
        is_child = group == MedicaidGroup.CHILD
        is_expansion_adult = group == MedicaidGroup.EXPANSION_ADULT
        is_non_expansion_adult = group == MedicaidGroup.NON_EXPANSION_ADULT

        # Calculate spend and enrollment for each group
        aged_spend = p.spending.by_eligibility_group.aged[state]
        disabled_spend = p.spending.by_eligibility_group.disabled[state]
        child_spend = p.spending.by_eligibility_group.child[state]
        expansion_adult_spend = (
            p.spending.by_eligibility_group.expansion_adults[state]
        )
        non_expansion_adult_spend = (
            p.spending.by_eligibility_group.non_expansion_adults[state]
        )

        aged_enroll = p.enrollment.aged[state]
        disabled_enroll = p.enrollment.disabled[state]
        child_enroll = p.enrollment.child[state]
        expansion_adult_enroll = p.enrollment.expansion_adults[state]
        non_expansion_adult_enroll = p.enrollment.non_expansion_adults[state]

        # Combine aged and disabled for AGED_DISABLED group
        aged_disabled_spend = aged_spend + disabled_spend
        aged_disabled_enroll = aged_enroll + disabled_enroll

        # Select the right values based on group
        spend = select(
            [
                is_aged_disabled,
                is_child,
                is_expansion_adult,
                is_non_expansion_adult,
            ],
            [
                aged_disabled_spend,
                child_spend,
                expansion_adult_spend,
                non_expansion_adult_spend,
            ],
            default=0,
        )

        enroll = select(
            [
                is_aged_disabled,
                is_child,
                is_expansion_adult,
                is_non_expansion_adult,
            ],
            [
                aged_disabled_enroll,
                child_enroll,
                expansion_adult_enroll,
                non_expansion_adult_enroll,
            ],
            default=0,
        )

        # Avoid divide‑by‑zero in non‑expansion states by assigning national average.
        per_capita = p.totals.per_capita[group]
        mask = enroll > 0
        per_capita[mask] = spend[mask] / enroll[mask]

        return per_capita



from policyengine_us.model_api import *


class tx_ccs_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Texas CCS income eligible"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/texas/40-Tex-Admin-Code-SS-809-41"
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.tx.twc.ccs
        # Get monthly countable income
        monthly_income = spm_unit("tx_ccs_countable_income", period)
        # Calculate monthly SMI limit
        smi = spm_unit("hhs_smi", period)
        income_limit = smi * p.income.smi_rate
        return monthly_income <= income_limit



from policyengine_us.model_api import *


class co_quality_rating_of_child_care_facility(Variable):
    value_type = int
    entity = Person
    label = "Quality rating of child care facility for Colorado Child Care Assistance Program"
    definition_period = YEAR
    # defined_for = StateCode.CO



from policyengine_us.model_api import *


class energy_efficient_home_improvement_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Energy efficient home improvement credit"
    documentation = "Residential clean energy credit"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/25C"

    def formula(tax_unit, period, parameters):
        credit_limit = tax_unit(
            "energy_efficient_home_improvement_credit_credit_limit", period
        )
        potential = tax_unit(
            "energy_efficient_home_improvement_credit_potential", period
        )
        return min_(credit_limit, potential)



from policyengine_us.model_api import *


class in_adoption_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana adoption exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/indiana/title-6/article-3/chapter-1/section-6-3-1-3-5/"  # (a)(5)(D)
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["in"].tax.income.exemptions
        person = tax_unit.members
        is_qualifying_dependent_child = person(
            "in_is_qualifying_dependent_child", period
        )
        adopted_this_year = person("adopted_this_year", period)
        is_qualifying_adopted_dependent_child = (
            is_qualifying_dependent_child & adopted_this_year
        )
        capped_exemption_amount = max_(
            p.adoption.amount - p.additional.amount, 0
        )
        return tax_unit.sum(
            is_qualifying_adopted_dependent_child * capped_exemption_amount
        )



from policyengine_us.model_api import *


class il_bcc_insurance_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois BCC insurance eligible"
    definition_period = YEAR
    reference = (
        "https://www.dhs.state.il.us/page.aspx?item=33528",
        "https://www.ilga.gov/legislation/ilcs/documents/030500050K5-2.htm",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        # Per PM 06-20-02, persons are eligible if they are "uninsured".
        # A person is NOT uninsured (thus ineligible) if they have:
        # - Medicaid eligibility through another pathway
        # - Health coverage that would pay for cancer treatment
        #
        # Exceptions that still qualify as "uninsured":
        # - Coverage with limited benefits
        # - Coverage where lifetime cap has been met/exceeded
        # - Coverage that excludes cancer treatment
        # - Indian Health Service coverage only
        on_medicaid = person("receives_medicaid", period)
        has_qualifying_coverage = person("has_bcc_qualifying_coverage", period)
        has_disqualifying_coverage = on_medicaid | has_qualifying_coverage
        return ~has_disqualifying_coverage



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_ut_refundable_eitc() -> Reform:
    class ut_has_qualifying_child_for_refundable_eitc(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Utah tax unit has qualifying child for refundable EITC"
        defined_for = StateCode.UT
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            person = tax_unit.members
            p = parameters(period).gov.contrib.states.ut.eitc
            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)
            is_qualifying_child = (age <= p.max_age) & is_dependent
            return tax_unit.any(is_qualifying_child)

    class ut_refundable_eitc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Utah refundable EITC"
        unit = USD
        definition_period = YEAR
        defined_for = "ut_has_qualifying_child_for_refundable_eitc"

        adds = ["ut_eitc"]

    class ut_eitc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Utah Earned Income Tax Credit"
        unit = USD
        documentation = "This credit is a fraction of the federal EITC."
        definition_period = YEAR
        defined_for = StateCode.UT
        reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1044.html?v=C59-10-S1044_2022050420220504"

        def formula(tax_unit, period, parameters):
            p = parameters(
                period
            ).gov.states.ut.tax.income.credits.earned_income
            federal_eitc = tax_unit("eitc", period)
            return p.rate * federal_eitc

    class ut_non_refundable_eitc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Utah non-refundable EITC"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.UT

        adds = ["ut_eitc"]
        subtracts = ["ut_refundable_eitc"]

    class ut_non_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        label = "Utah non-refundable tax credits"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.UT

        adds = [
            "ut_non_refundable_eitc",
            "ut_retirement_credit",
            "ut_ss_benefits_credit",
            "ut_at_home_parent_credit",
            "ut_ctc",
        ]

    class ut_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        label = "Utah refundable credits"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.UT

        adds = ["ut_refundable_eitc"]

    class reform(Reform):
        def apply(self):
            self.update_variable(ut_has_qualifying_child_for_refundable_eitc)
            self.update_variable(ut_refundable_eitc)
            self.update_variable(ut_non_refundable_eitc)
            self.update_variable(ut_eitc)
            self.update_variable(ut_non_refundable_credits)
            self.update_variable(ut_refundable_credits)

    return reform


def create_ut_refundable_eitc_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_ut_refundable_eitc()

    p = parameters.gov.contrib.states.ut.eitc

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_ut_refundable_eitc()
    else:
        return None


ut_refundable_eitc = create_ut_refundable_eitc_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class amt_part_iii_required(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax Part III required"
    documentation = "Whether the Alternative Minimum Tax (AMT) Part III worksheet is required, Form 6251, Part III"
    reference = "https://www.irs.gov/pub/irs-pdf/f6251.pdf"

    def formula(tax_unit, period, parameters):
        relevant_inputs = add(
            tax_unit,
            period,
            [
                "dwks10",
                "dwks13",
                "dwks14",
                "dwks19",
                "unrecaptured_section_1250_gain",
            ],
        )
        return relevant_inputs > 0



from policyengine_us.model_api import *


class is_wic_at_nutritional_risk(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    documentation = "Assessed as being at nutritional risk for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "At nutritional risk for WIC"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#b_8"

    def formula(person, period, parameters):
        # Assign nutritional risk status probabilistically in microsimulation.
        # Nutritional risk is a health assessment independent of income per § 1786(b)(8).
        # Assume all meet qualification in individual simulation.
        if person.simulation.dataset is not None:
            wic_reported = person("receives_wic", period)
            category = person("wic_category", period)
            risk = parameters(period).gov.usda.wic.nutritional_risk
            imputed_risk = random(person) < risk[category]
            return wic_reported | imputed_risk
        else:
            return True



from policyengine_us.model_api import *


class ma_tafdc_potential_main_benefit(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) potential main benefit"
    definition_period = MONTH
    reference = "https://www.mass.gov/how-to/transitional-aid-to-families-with-dependent-children-tafdc"
    defined_for = "ma_tafdc_eligible"

    def formula(spm_unit, period, parameters):
        payment_standard = spm_unit("ma_tafdc_payment_standard", period)
        countable_income = spm_unit(
            "ma_tafdc_applicable_income_grant_amount", period
        )
        return max_(0, payment_standard - countable_income)



from policyengine_us.model_api import *


class tx_tanf_categorically_eligible_person(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Categorically eligible person for Texas TANF"
    reference = (
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-201",
        "https://www.law.cornell.edu/regulations/texas/1-Tex-Admin-Code-SS-372-104",
    )
    defined_for = StateCode.TX

    def formula(person, period, parameters):
        immigration_eligible = person("is_citizen_or_legal_immigrant", period)
        ssi = person("ssi", period)

        return immigration_eligible & (ssi == 0)



from policyengine_us.model_api import *


class housing_designated_welfare(Variable):
    value_type = float
    entity = SPMUnit
    label = "Housing designated welfare"
    unit = USD
    documentation = "Housing designated welfare"
    definition_period = YEAR



from policyengine_us.model_api import *


class income_tax_main_rates(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Income tax main rates"
    reference = "https://www.law.cornell.edu/uscode/text/26/1"
    unit = USD

    def formula(tax_unit, period, parameters):
        # compute taxable income that is taxed at the main rates
        full_taxable_income = tax_unit("taxable_income", period)
        cg_exclusion = tax_unit(
            "capital_gains_excluded_from_taxable_income", period
        )
        taxinc = max_(0, full_taxable_income - cg_exclusion)
        # compute tax using bracket rates and thresholds
        p = parameters(period).gov.irs.income
        bracket_tops = p.bracket.thresholds
        bracket_rates = p.bracket.rates
        filing_status = tax_unit("filing_status", period)
        tax = 0
        bracket_bottom = 0
        for i in range(1, len(list(bracket_rates.__iter__())) + 1):
            b = str(i)
            bracket_top = bracket_tops[b][filing_status]
            tax += bracket_rates[b] * amount_between(
                taxinc, bracket_bottom, bracket_top
            )
            bracket_bottom = bracket_top
        return tax



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_ri_ctc() -> Reform:
    class ri_ctc_eligible_children(Variable):
        value_type = int
        entity = TaxUnit
        label = "Rhode Island CTC eligible children count"
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            person = tax_unit.members
            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)

            # Check age eligibility
            meets_age = age < p.age_limit

            eligible = is_dependent & meets_age
            return tax_unit.sum(eligible)

    class ri_ctc_young_child_boost(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island CTC young child boost"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            person = tax_unit.members
            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)

            # Check both general CTC eligibility and young child age limit
            meets_age = age < p.age_limit
            meets_young_child_age = age < p.young_child_boost.age_limit

            eligible_young_children = tax_unit.sum(
                is_dependent & meets_age & meets_young_child_age
            )
            return eligible_young_children * p.young_child_boost.amount

    class ri_ctc_maximum(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island CTC maximum amount before phaseout"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            # Base credit for all eligible children
            eligible_children = tax_unit("ri_ctc_eligible_children", period)
            base_credit = eligible_children * p.amount

            # Young child boost
            young_child_boost = tax_unit("ri_ctc_young_child_boost", period)

            return base_credit + young_child_boost

    class ri_ctc_phaseout(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island CTC phaseout amount"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            filing_status = tax_unit("filing_status", period)
            agi = tax_unit("ri_agi", period)

            threshold = p.phaseout.threshold[filing_status]
            excess_income = max_(agi - threshold, 0)
            return excess_income * p.phaseout.rate

    class ri_total_ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island Child Tax Credit"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            # Calculate maximum credit
            maximum = tax_unit("ri_ctc_maximum", period)

            # Apply phaseout
            phaseout = tax_unit("ri_ctc_phaseout", period)

            return max_(maximum - phaseout, 0)

    class ri_refundable_ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island CTC refundable portion"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            total_credit = tax_unit("ri_total_ctc", period)

            # The refundable portion is the minimum of the cap and total credit
            # This ensures the refundable amount never exceeds the total credit
            # - If cap = 0: credit is nonrefundable
            # - If 0 < cap < total_credit: credit is partially refundable
            # - If cap >= total_credit: credit is fully refundable
            return min_(total_credit, p.refundability.cap)

    class ri_non_refundable_ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Rhode Island CTC non-refundable portion"
        unit = USD
        definition_period = YEAR
        defined_for = StateCode.RI

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.contrib.states.ri.ctc

            total_credit = tax_unit("ri_total_ctc", period)
            refundable_portion = tax_unit("ri_refundable_ctc", period)
            return max_(total_credit - refundable_portion, 0)

    def modify_parameters(parameters):
        # Add ri_refundable_ctc to refundable credits list
        refundable = parameters.gov.states.ri.tax.income.credits.refundable
        current_refundable = refundable(instant("2025-01-01"))
        if "ri_refundable_ctc" not in current_refundable:
            new_refundable = list(current_refundable) + ["ri_refundable_ctc"]
            refundable.update(
                start=instant("2025-01-01"),
                stop=instant("2100-12-31"),
                value=new_refundable,
            )

        # Add ri_non_refundable_ctc to non-refundable credits list
        non_refundable = (
            parameters.gov.states.ri.tax.income.credits.non_refundable
        )
        current_non_refundable = non_refundable(instant("2025-01-01"))
        if "ri_non_refundable_ctc" not in current_non_refundable:
            new_non_refundable = list(current_non_refundable) + [
                "ri_non_refundable_ctc"
            ]
            non_refundable.update(
                start=instant("2025-01-01"),
                stop=instant("2100-12-31"),
                value=new_non_refundable,
            )

        return parameters

    class reform(Reform):
        def apply(self):
            self.update_variable(ri_ctc_eligible_children)
            self.update_variable(ri_ctc_young_child_boost)
            self.update_variable(ri_ctc_maximum)
            self.update_variable(ri_ctc_phaseout)
            self.update_variable(ri_total_ctc)
            self.update_variable(ri_refundable_ctc)
            self.update_variable(ri_non_refundable_ctc)
            self.modify_parameters(modify_parameters)

    return reform


def create_ri_ctc_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_ri_ctc()

    p = parameters.gov.contrib.states.ri.ctc

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_ri_ctc()
    else:
        return None


ri_ctc = create_ri_ctc_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class ny_pension_exclusion(Variable):
    value_type = float
    entity = Person
    label = "New York pension exclusion"
    unit = USD
    documentation = "Exclusion for pension income for eligible individuals."
    definition_period = YEAR
    dict(
        title="N.Y. Comp. Codes R. & Regs. tit. 20 § 112.3",
        href="https://casetext.com/regulation/new-york-codes-rules-and-regulations/title-20-department-of-taxation-and-finance/chapter-ii-income-taxes-and-estate-taxes/subchapter-a-new-york-state-personal-income-tax-under-article-22-of-the-tax-law/article-2-residents/part-112-new-york-adjusted-gross-income-of-a-resident-individual/section-1123-modifications-reducing-federal-adjusted-gross-income",  # (c)
    )

    def formula(person, period, parameters):
        pension_income = person("taxable_pension_income", period)
        age = person("age", period)

        # Fetching values from separate YAML files
        p = parameters(
            period
        ).gov.states.ny.tax.income.agi.subtractions.pension_exclusion

        meets_age_test = age >= p.min_age

        return meets_age_test * min_(pension_income, p.cap)



from policyengine_us.model_api import *


class nj_gross_income(Variable):
    value_type = float
    entity = Person
    label = "New Jersey gross income"
    unit = USD
    documentation = "Gross income calculated from specific income categories per NJ statute, before additions and subtractions. This is built from gross income sources, not federal AGI."
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/new-jersey/title-54a/section-54a-5-1/",
        "https://www.nj.gov/treasury/taxation/pdf/current/1040.pdf",  # Lines 15-27
    )
    defined_for = StateCode.NJ

    adds = "gov.states.nj.tax.income.gross_income_sources"



from policyengine_us.model_api import *


class itemized_taxable_income_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Itemized taxable income deductions"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        total_deductions = tax_unit(
            "total_itemized_taxable_income_deductions", period
        )
        reduction = tax_unit(
            "itemized_taxable_income_deductions_reduction", period
        )
        return max_(0, total_deductions - reduction)



from policyengine_us.model_api import *


class veterans_benefits(Variable):
    value_type = float
    entity = Person
    label = "Veterans benefits"
    unit = USD
    documentation = "Veterans benefits from past military service."
    definition_period = YEAR



from policyengine_us.model_api import *


class il_aabd_is_bedfast(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Whether the person is a bedfast client under the Illinois Aid to the Aged, Blind or Disabled (AABD)"
    reference = "https://www.dhs.state.il.us/page.aspx?item=15913"
    defined_for = StateCode.IL



from policyengine_us.model_api import *


class ia_is_tax_exempt(Variable):
    value_type = bool
    entity = TaxUnit
    label = "whether or not exempt from Iowa income tax because of low income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=37"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=37"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        is_single = filing_status == filing_status.possible_values.SINGLE
        p = parameters(period).gov.states.ia.tax.income.tax_exempt
        elderly_head = tax_unit("age_head", period) >= p.elderly_age
        elderly_spouse = tax_unit("age_spouse", period) >= p.elderly_age
        is_elderly = elderly_head | elderly_spouse
        pil = p.income_limit
        modified_income_limit = where(
            is_single,
            where(is_elderly, pil.single_elderly, pil.single_nonelderly),
            where(is_elderly, pil.other_elderly, pil.other_nonelderly),
        )
        return tax_unit("ia_modified_income", period) <= modified_income_limit



from policyengine_us.model_api import *


class nd_rtrc(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Dakota resident-tax-relief nonrefundable credit amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2021-iit/form-nd-1-2021.pdf#page=2"  # line 20
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2021-iit/individual-income-tax-booklet-2021.pdf#page=14"
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/form-nd-1-2022.pdf#page=2"  # line 20
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf#page=14"
    )
    defined_for = StateCode.ND

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nd.tax.income.credits
        filing_status = tax_unit("filing_status", period)
        rtrc_amount = where(
            filing_status == filing_status.possible_values.JOINT,
            p.resident_tax_relief.joint_amount,
            p.resident_tax_relief.other_amount,
        )
        inctax = tax_unit("nd_income_tax_before_credits", period)
        return min_(rtrc_amount, inctax)



from policyengine_us.model_api import *


def create_abolish_payroll_tax() -> Reform:
    class household_tax_before_refundable_credits(Variable):
        value_type = float
        entity = Household
        label = "total tax before refundable credits"
        documentation = "Total tax liability before refundable credits."
        unit = USD
        definition_period = YEAR

        def formula(household, period, parameters):
            p = parameters(period)
            added_components = (
                p.gov.household.household_tax_before_refundable_credits
            )
            added_components = [
                c for c in added_components if c != "employee_payroll_tax"
            ]
            return add(household, period, added_components)

    class reform(Reform):
        def apply(self):
            self.update_variable(household_tax_before_refundable_credits)

    return reform


def create_abolish_payroll_tax_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_abolish_payroll_tax()

    p = parameters(period).gov.contrib.ubi_center.flat_tax

    if p.abolish_payroll_tax:
        return create_abolish_payroll_tax()
    else:
        return None


abolish_payroll_tax = create_abolish_payroll_tax_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class wv_low_income_family_tax_credit_fpg(Variable):
    value_type = float
    entity = TaxUnit
    label = "Federal poverty guidelines for the West Virginia low-income family tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = "wv_low_income_family_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.wv.tax.income.credits.liftc  # low_income_family_tax_credit

        # max family size limit
        n = tax_unit("tax_unit_size", period)
        state_group = tax_unit.household("state_group_str", period)

        p_fpg = parameters(period).gov.hhs.fpg
        p1 = p_fpg.first_person[state_group]
        pn = p_fpg.additional_person[state_group]
        family_size = min_(n, p.max_family_size)
        return p1 + pn * (family_size - 1)



from policyengine_us.model_api import *


class SnapUtilityRegion(Enum):
    AL = "AL"
    AK = "AK"
    AZ = "AZ"
    AR = "AR"
    CA = "CA"
    CO = "CO"
    CT = "CT"
    DE = "DE"
    FL = "FL"
    GA = "GA"
    HI = "HI"
    ID = "ID"
    IL = "IL"
    IN = "IN"
    IA = "IA"
    KS = "KS"
    KY = "KY"
    LA = "LA"
    ME = "ME"
    MD = "MD"
    MA = "MA"
    MI = "MI"
    MN = "MN"
    MS = "MS"
    MO = "MO"
    MT = "MT"
    NE = "NE"
    NV = "NV"
    NH = "NH"
    NJ = "NJ"
    NM = "NM"
    NY = "NY"
    NC = "NC"
    ND = "ND"
    OH = "OH"
    OK = "OK"
    OR = "OR"
    PA = "PA"
    RI = "RI"
    SC = "SC"
    SD = "SD"
    TN = "TN"
    TX = "TX"
    UT = "UT"
    VT = "VT"
    VA = "VA"
    WA = "WA"
    WV = "WV"
    WI = "WI"
    WY = "WY"
    DC = "DC"
    GU = "GU"
    MP = "MP"
    PW = "PW"
    PR = "PR"
    VI = "VI"
    AA = "AA"
    AE = "AE"
    AP = "AP"


class snap_utility_region(Variable):
    value_type = Enum
    possible_values = SnapUtilityRegion
    default_value = SnapUtilityRegion.CA
    entity = Household
    label = "SNAP utility region"
    documentation = "Region deciding the SNAP utility allowances."
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("state_code", period).decode_to_str()



from policyengine_us.model_api import *


class nc_scca_has_eligible_child(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Any eligible child for North Carolina Subsidized Child Care Assistance program"
    definition_period = YEAR
    defined_for = StateCode.NC
    reference = "https://ncchildcare.ncdhhs.gov/Portals/0/documents/pdf/A/ACF-118_CCDF_FFY_2022-2024_For_North_Carolina_Amendment_1.pdf?ver=C9YfIUPAFekeBA3I1mN8aA%3d%3d#page=83"
    adds = ["nc_scca_child_age_eligible"]



from policyengine_us.model_api import *


class nj_additions(Variable):
    value_type = float
    entity = Person
    label = "New Jersey additions to gross income"
    unit = USD
    documentation = "Additions to New Jersey gross income per NJ Statute 54A:5-1. These are amounts added back to gross income that may have been excluded or deducted for federal purposes."
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/new-jersey/title-54a/section-54a-5-1/",
        "https://www.nj.gov/treasury/taxation/pdf/current/1040.pdf",  # Lines 28-38
    )
    defined_for = StateCode.NJ

    adds = "gov.states.nj.tax.income.additions"



from policyengine_us.model_api import *


class fsla_overtime_premium(Variable):
    value_type = float
    entity = Person
    label = "premium income from overtime hours worked"
    unit = USD
    definition_period = YEAR
    defined_for = "is_eligible_for_fsla_overtime"

    def formula_2014(person, period, parameters):
        p = parameters(period).gov.irs.income.exemption.overtime
        worked_hours = person("hours_worked_last_week", period)

        weekly_overtime_hours = max_(worked_hours - p.hours_threshold, 0)
        # Straight-time–equivalent hours in the year
        hour_equivalents = WEEKS_IN_YEAR * (
            p.hours_threshold + weekly_overtime_hours * p.rate_multiplier
        )
        # Implied straight-time hourly wage
        base_rate = person("employment_income", period) / hour_equivalents
        # Return only the overtime premium (e.g. the 50 % in time-and-a-half)
        return (
            base_rate
            * (p.rate_multiplier - 1)
            * weekly_overtime_hours
            * WEEKS_IN_YEAR
        )



from policyengine_us.model_api import *


class il_tanf_gross_earned_income(Variable):
    value_type = float
    entity = Person
    label = "Illinois Temporary Assistance for Needy Families (TANF) gross earned income"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.IL

    adds = "gov.states.il.dhs.tanf.income.sources.earned"



from policyengine_us.model_api import *


class snap_net_income_pre_shelter(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "SNAP net income before the shelter deduction, needed as intermediate to calculate shelter deduction"
    label = "SNAP net income (pre-shelter)"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_6_A"

    def formula(spm_unit, period, parameters):
        """
        A household shall be entitled, with respect to expenses other than
        expenses paid on behalf of the household by a third party, to an excess
        shelter expense deduction to the extent that the monthly amount
        expended by a household for shelter exceeds an amount equal to 50
        percent of monthly household income after all other applicable
        deductions have been allowed.
        """
        income = spm_unit("snap_gross_income", period)
        all_deductions = parameters(
            period
        ).gov.usda.snap.income.deductions.allowed
        deductions_except_shelter = [
            deduction
            for deduction in all_deductions
            if deduction != "snap_excess_shelter_expense_deduction"
        ]
        deduction_value = add(spm_unit, period, deductions_except_shelter)
        return max_(income - deduction_value, 0)



from policyengine_us.model_api import *


class tx_tanf_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas TANF countable income"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits",
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1350-calculating-household-income",
    )
    defined_for = StateCode.TX

    adds = [
        "tx_tanf_countable_earned_income",
        "tx_tanf_countable_unearned_income",
    ]



from policyengine_us.model_api import *


class ma_ccfa(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = (
        "Massachusetts Child Care Financial Assistance (CCFA) benefit amount"
    )
    definition_period = MONTH
    defined_for = "ma_ccfa_eligible"
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"

    def formula(spm_unit, period, parameters):
        pre_subsidy_childcare_expenses = spm_unit(
            "spm_unit_pre_subsidy_childcare_expenses", period.this_year
        )
        copay = spm_unit("ma_ccfa_total_copay", period)
        max_reimbursement = add(
            spm_unit, period, ["ma_ccfa_maximum_reimbursement"]
        )
        uncapped_benefit = max_(pre_subsidy_childcare_expenses - copay, 0)

        return min_(uncapped_benefit, max_reimbursement)



from policyengine_us.model_api import *


class il_tanf_payment_eligible_requirements(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible requirements for Illinois Temporary Assistance for Needy Families (TANF) payment"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.10",
        "https://www.dhs.state.il.us/page.aspx?item=13251",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        immigration_status_eligible = person(
            "il_tanf_immigration_status_eligible_person", period
        )
        ssi = person("ssi", period)
        receives_ssi = ssi > 0

        return ~receives_ssi & immigration_status_eligible



from policyengine_us.model_api import *


class ar_personal_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas personal credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2021_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf#page=1"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=12"
    )
    defined_for = StateCode.AR

    adds = [
        "ar_personal_credit_dependent",
        "ar_personal_credits_base",
        "ar_personal_credit_disabled_dependent",
    ]



from policyengine_us.model_api import *


class ca_riv_general_relief_meets_work_requirements(Variable):
    value_type = bool
    entity = Person
    label = "Meets work requirements for the Riverside County General Relief"
    definition_period = MONTH
    defined_for = "in_riv"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.local.ca.riv.general_relief.work_exempted_age
        # Person who is actively searching for jobs also qualify for work requirements
        is_working = (
            add(
                person, period, ["employment_income", "self_employment_income"]
            )
            > 0
        )
        # Check if person is a qualifying secondary school student
        age = person("monthly_age", period)
        is_full_time_student = person("is_full_time_student", period)
        in_secondary_school = person("is_in_secondary_school", period)
        is_qualifying_secondary_student = (
            is_full_time_student & in_secondary_school
        )
        # Age-based exemptions have different thresholds based on student status:
        # - Full-time secondary students: exempt if younger than threshold
        # - Non-students: exempt if older than threshold (senior citizens)
        work_exempted_due_to_age = where(
            is_qualifying_secondary_student,
            age < p.younger,  # Student exemption: too young
            age >= p.older,  # Senior exemption: too old
        )
        # Incapacitated or need to caring full time for a member of household
        has_incapable_of_self_care = person.spm_unit.any(
            person("is_incapable_of_self_care", period)
        )
        return (
            is_working | work_exempted_due_to_age | has_incapable_of_self_care
        )



from policyengine_us.model_api import *


class co_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado child tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        # C.R.S. 39-22-129. Child tax credit - legislative declaration - definitions.
        "https://casetext.com/statute/colorado-revised-statutes/title-39-taxation/specific-taxes/income-tax/article-22-income-tax/part-1-general/section-39-22-129-child-tax-credit-legislative-declaration-definitions-repeal",
        # 2022 Colorado Child Tax Credit
        "https://tax.colorado.gov/sites/tax/files/documents/DR_0104CN_2022.pdf#page=1",
        # Colorado Individual Income Tax Filing Guide - Instructions for Select Credits from the DR 0104CR - Line 1 Child Tax Credit
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=16",
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.co.tax.income.credits.ctc
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        person = tax_unit.members
        # Depending on the year, Colorado has based its CTC on AGI, the federal CTC, and eligible children.
        agi = tax_unit("adjusted_gross_income", period)
        if p.ctc_matched_federal_credit:
            federal_ctc = tax_unit("co_federal_ctc", period)
            rate = select(
                [
                    filing_status == statuses.SINGLE,
                    filing_status == statuses.JOINT,
                    filing_status == statuses.SEPARATE,
                    filing_status == statuses.SURVIVING_SPOUSE,
                    filing_status == statuses.HEAD_OF_HOUSEHOLD,
                ],
                [
                    p.rate.single.calc(agi, right=True),
                    p.rate.joint.calc(agi, right=True),
                    p.rate.separate.calc(agi, right=True),
                    p.rate.surviving_spouse.calc(agi, right=True),
                    p.rate.head_of_household.calc(agi, right=True),
                ],
            )
            return rate * federal_ctc
        else:
            person = tax_unit.members
            child_age_eligible = person("age", period) < p.age_threshold

            eligible_child = (
                person("co_ctc_eligible_child", period) & child_age_eligible
            )
            eligible_children = tax_unit.sum(eligible_child)
            amount_per_child = select(
                [
                    filing_status == statuses.SINGLE,
                    filing_status == statuses.JOINT,
                    filing_status == statuses.SEPARATE,
                    filing_status == statuses.SURVIVING_SPOUSE,
                    filing_status == statuses.HEAD_OF_HOUSEHOLD,
                ],
                [
                    p.amount.single.calc(agi, right=True),
                    p.amount.joint.calc(agi, right=True),
                    p.amount.separate.calc(agi, right=True),
                    p.amount.surviving_spouse.calc(agi, right=True),
                    p.amount.head_of_household.calc(agi, right=True),
                ],
            )
            return amount_per_child * eligible_children



from policyengine_us.model_api import *


class taxsim_depx(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of dependents"
    unit = "person"
    documentation = "Number of dependents (for personal exemption calculation). In the case of a file submission, if no age1 or dep19 variable is present, depx is used for the number of eligible children. You can negate this assumption by putting a large number (such as 99) in the age1 field."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_dependent = person("is_tax_unit_dependent", period)
        return tax_unit.sum(is_dependent)



from policyengine_us.model_api import *


class ut_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah taxable income"
    unit = USD
    documentation = "Form TC-40, line 9"
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = "https://tax.utah.gov/forms/2021/tc-40.pdf#page=1"

    adds = ["ut_total_income"]
    subtracts = ["ut_subtractions", "ut_state_tax_refund"]



from policyengine_us.model_api import *


class ia_reportable_social_security(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa reportable social security benefits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=11"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=10"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        benefits = add(tax_unit, period, ["social_security"])
        income = (
            tax_unit("adjusted_gross_income", period)
            - add(tax_unit, period, ["taxable_social_security"])
            + add(tax_unit, period, ["tax_exempt_interest_income"])
        )
        p = parameters(period).gov.states.ia.tax.income
        filing_status = tax_unit("filing_status", period)
        deduction = p.reportable_social_security.deduction[filing_status]
        return p.reportable_social_security.fraction * min_(
            benefits, max_(0, income - deduction)
        )



from policyengine_us.model_api import *


class or_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=18",
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # Subsection 316.037
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        income = tax_unit("or_taxable_income", period)
        rates = parameters(period).gov.states["or"].tax.income.rates
        return select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.JOINT,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.SURVIVING_SPOUSE,
            ],
            [
                rates.single.calc(income),
                rates.joint.calc(income),
                rates.head_of_household.calc(income),
                rates.separate.calc(income),
                rates.surviving_spouse.calc(income),
            ],
        )



from policyengine_us.model_api import *


class ca_federal_capped_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Capped child/dependent care credit replicated to include California limitations"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/2020/2020-3506-instructions.html"
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        cdcc = tax_unit("ca_federal_cdcc", period)
        p = parameters(period).gov.irs.credits
        # follow Credit Limit Worksheet in 2022 Form 2441 instructions:
        itaxbc = tax_unit("income_tax_before_credits", period)  # WS Line1
        # Excess Advance PTC Repayment (Form 8962) assumed zero in above line
        offset = tax_unit("foreign_tax_credit", period)
        # Partner Additional Reporting Year Tax (Form 8978) assumed zero above
        cap = max_(itaxbc - offset, 0)  # WS Line 2
        return min_(cdcc, cap)  # WS Line 3



from policyengine_us.model_api import *


class qualified_business_income_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Qualified business income deduction for tax unit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/199A#b_1"
        "https://www.irs.gov/pub/irs-prior/p535--2018.pdf"
    )

    def formula(tax_unit, period, parameters):
        # compute sum of QBID amounts for each person in TaxUnit following
        # logic in 2018 IRS Publication 535, Worksheet 12-A, line 16
        person = tax_unit.members
        qbid_amt = person("qbid_amount", period)
        uncapped_qbid = tax_unit.sum(qbid_amt)
        # apply taxinc cap at the TaxUnit level following logic
        # in 2018 IRS Publication 535, Worksheet 12-A, lines 32-37
        taxinc_less_qbid = tax_unit("taxable_income_less_qbid", period)
        netcg_qdiv = tax_unit("adjusted_net_capital_gain", period)
        p = parameters(period).gov.irs.deductions.qbi
        taxinc_cap = p.max.rate * max_(0, taxinc_less_qbid - netcg_qdiv)
        pre_floor_qbid = min_(uncapped_qbid, taxinc_cap)
        if p.deduction_floor.in_effect:
            qualified_business_income = tax_unit(
                "qualified_business_income", period
            )
            floor = p.deduction_floor.amount.calc(qualified_business_income)
            return max_(pre_floor_qbid, floor)
        return pre_floor_qbid



from policyengine_us.model_api import *


class mo_pension_and_ss_or_ssd_deduction_section_b(Variable):
    value_type = float
    entity = Person
    label = "Missouri Pension and Social Security or SS Disability Deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf#page=3",
        "https://dor.mo.gov/forms/MO-1040%20Fillable%20Calculating_2021.pdf#page=2",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.124",
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        tax_unit = person.tax_unit
        ind_mo_agi = person("mo_adjusted_gross_income", period)
        unit_mo_agi = tax_unit.sum(ind_mo_agi)
        ind_taxable_oasdi = person("taxable_social_security", period)
        unit_taxable_oasdi = tax_unit.sum(ind_taxable_oasdi)
        p = parameters(period).gov.states.mo.tax.income.deductions
        filing_status = tax_unit("filing_status", period)
        excess_agi = max_(  # line5
            0,
            (
                unit_mo_agi
                - unit_taxable_oasdi
                - p.mo_private_pension_deduction_allowance[filing_status]
            ),
        )
        ind_pvt_pen_amt = person("taxable_private_pension_income", period)
        ind_pvt_pen_val = min_(ind_pvt_pen_amt, p.mo_max_private_pension)
        unit_pvt_pen_val = tax_unit.sum(ind_pvt_pen_val)  # line8
        unit_deduction = max_(0, unit_pvt_pen_val - excess_agi)  # line9
        # Compute the individual's share of the tax unit's taxable private pension income.
        # Use a mask rather than where to avoid a divide-by-zero warning. Default to zero.
        ind_share = np.zeros_like(ind_pvt_pen_val)
        mask = ind_pvt_pen_val > 0
        ind_share[mask] = ind_pvt_pen_val[mask] / unit_pvt_pen_val[mask]

        return ind_share * unit_deduction



from policyengine_us.model_api import *


class ca_riv_general_relief_special_needs_amount(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Riverside County General Relief special needs amount"
    definition_period = YEAR
    defined_for = "in_riv"

    adds = [
        "gov.local.ca.riv.general_relief.needs_standards.special_needs.transportation"
    ]



from policyengine_us.model_api import *


class tx_tanf_gross_unearned_income(Variable):
    value_type = float
    entity = Person
    label = "Texas TANF gross unearned income"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits",
        "https://www.law.cornell.edu/regulations/texas/1-TAC-372-601",
    )
    defined_for = StateCode.TX

    adds = "gov.states.tx.tanf.income.sources.unearned"



from policyengine_us.model_api import *


class pr_earned_income_credit_unearned_income(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Puerto Rico earned income credit unearned income"
    definition_period = YEAR
    reference = "https://casetext.com/statute/laws-of-puerto-rico/title-thirteen-taxation-and-finance/subtitle-17-internal-revenue-code-of-2011/part-ii-income-taxes/chapter-1007-credits-against-tax/subchapter-b-refundable-credits/30211-earned-income-credit"

    adds = "gov.territories.pr.tax.income.credits.earned_income.unearned_income.sources"



from policyengine_us.model_api import *


class mi_homestead_property_tax_credit_pre_alternate_senior_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan homestead property tax credit per alternate senior credit amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-508",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040CR.pdf#page=2",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=34",
    )
    defined_for = "mi_homestead_property_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.credits.homestead_property_tax.reduction
        # Line 42
        total_household_resources = tax_unit("mi_household_resources", period)
        homestead_allowable_credit = tax_unit(
            "mi_allowable_homestead_property_tax_credit", period
        )
        # Line 43
        # The reduction is specified as going from 100% to 0% rather than vice-versa.
        excess = max_(total_household_resources - p.start, 0)
        increments = np.ceil(excess / p.increment)
        phase_out_rate = max_(1 - increments * p.rate, 0)
        # Line 44
        return phase_out_rate * homestead_allowable_credit



from policyengine_us.model_api import *


class snap(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "Final SNAP benefit amount, equal to net income minus food contribution"
    label = "SNAP allotment"
    reference = "https://www.law.cornell.edu/uscode/text/7/2017#a"
    unit = USD
    exhaustive_parameter_dependencies = [
        "gov.usda.snap",
        "gov.ssa",
    ]

    def formula(spm_unit, period, parameters):
        takes_up = spm_unit("takes_up_snap_if_eligible", period)
        is_in_microsim = hasattr(spm_unit.simulation, "dataset")
        if parameters(period).gov.usda.snap.abolish_snap:
            return 0
        elif parameters(period).gov.simulation.reported_snap:
            return spm_unit("snap_reported", period)
        else:
            value = add(
                spm_unit,
                period,
                [
                    "snap_normal_allotment",
                    "snap_emergency_allotment",
                    "dc_snap_temporary_local_benefit",
                ],
            )
            if is_in_microsim:
                return value * takes_up
            else:
                return value



from policyengine_us.model_api import *


class amt_separate_addition(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "AMT taxable income separate addition"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/55#b_2"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("taxable_income", period)
        excluded_deductions = tax_unit("amt_excluded_deductions", period)
        amt_inc = taxable_income + excluded_deductions
        p = parameters(period).gov.irs.income.amt.exemption
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        reduced_amt_inc = max_(0, amt_inc - p.separate_limit)
        return (
            max_(
                0,
                min_(
                    p.amount[filing_status],
                    p.phase_out.rate * reduced_amt_inc,
                ),
            )
            * separate
        )



from policyengine_us.model_api import *


class mi_standard_home_heating_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Michigan home heating standard credit"
    definition_period = YEAR
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.credits.home_heating.standard

        mi_household_resources = tax_unit("mi_household_resources", period)
        # the poverty guidelines are determined by the number of exemptions
        # which is a function of the tax_unit_size which is used in tax_unit_fpg
        fpg = tax_unit("tax_unit_fpg", period)
        limit = fpg * p.fpg_rate
        return mi_household_resources <= limit



from policyengine_us.model_api import *


class ca_ala_general_assistance_immigration_status_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Alameda County General Assistance due to immigration status"
    definition_period = MONTH
    defined_for = "in_ala"
    reference = "https://www.alamedacountysocialservices.org/acssa-assets/PDF/GA-Policies/GA-Regulations.pdf#page=21"

    def formula(person, period, parameters):
        p = parameters(period).gov.local.ca.ala.general_assistance
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        return np.isin(immigration_status_str, p.qualified_immigration_status)



from policyengine_us.model_api import *


class ut_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah Earned Income Tax Credit"
    unit = USD
    documentation = "This credit is a fraction of the federal EITC."
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1044.html?v=C59-10-S1044_2022050420220504"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ut.tax.income.credits.earned_income
        federal_eitc = tax_unit("eitc", period)
        return p.rate * federal_eitc



from policyengine_us.model_api import *


class pr_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico deductions"
    unit = USD
    reference = "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-c/30135/"
    definition_period = YEAR
    defined_for = StateCode.PR

    adds = "gov.territories.pr.tax.income.taxable_income.deductions.sources"



from policyengine_us.model_api import *


class ca_state_supplement_aged_disabled_amount(Variable):
    value_type = float
    entity = SPMUnit
    label = "California SSI state supplement aged disabled amount"
    unit = USD
    definition_period = MONTH
    defined_for = "ca_state_supplement_aged_disabled_count"
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ca.cdss.state_supplement.payment_standard.aged_or_disabled.amount
        # Aged or disabled amount
        aged_disabled_count = spm_unit(
            "ca_state_supplement_aged_disabled_count", period
        )
        return select(
            [aged_disabled_count > 1, aged_disabled_count == 1],
            [p.married, p.single],
            default=0,
        )



from policyengine_us.model_api import *


class tx_tanf_countable_earned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas TANF countable earned income"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-1340-income-limits",
        "https://www.law.cornell.edu/regulations/texas/1-TAC-372-605",
    )
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        # Sum person-level earned income after work expense and disregards
        earned_after_disregards = spm_unit(
            "tx_tanf_earned_income_after_disregard_person", period
        )

        # Apply dependent care deduction (per § 372.409 (a)(3))
        dependent_care_deduction = spm_unit(
            "tx_tanf_dependent_care_deduction", period
        )

        # Countable earned income after all deductions
        return max_(earned_after_disregards - dependent_care_deduction, 0)



from policyengine_us.model_api import *


class mn_cdcc_dependent_count(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota child and dependent care expense credit dependent count"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1cd_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1cd_22_0.pdf"
    )
    defined_for = "mn_cdcc_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.credits.cdcc
        person = tax_unit.members
        # calculate number of qualifying dependents
        # ... children
        age = person("age", period)
        qualifies_by_age = age < p.child_age
        # ... disability
        non_head = ~person("is_tax_unit_head", period)
        disabled = person("is_incapable_of_self_care", period)
        qualifies_by_disability = non_head & disabled
        return tax_unit.sum(qualifies_by_age | qualifies_by_disability)



from policyengine_us.model_api import *


class ma_eaedc_net_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts EAEDC net income"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-500"  # (B) step 2

    adds = [
        "ma_tcap_gross_unearned_income",
        "ma_eaedc_countable_earned_income",
    ]



from policyengine_us.model_api import *


class spm_unit_net_income_reported(Variable):
    value_type = float
    entity = SPMUnit
    label = "Reported net income"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class nj_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-4-17-1/"
        "https://www.nj.gov/treasury/taxation/pdf/other_forms/tgi-ee/2021/1040i.pdf#page=44"
        "https://www.nj.gov/treasury/taxation/pdf/other_forms/tgi-ee/2022/1040i.pdf#page=44"
        "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=46"
    )
    defined_for = "nj_ctc_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.nj.tax.income.credits.ctc

        # Get amount per qualifying child based on taxable income.
        taxable_income = tax_unit("nj_taxable_income", period)
        amount_per_qualifying_child = p.amount.calc(taxable_income)

        # Get the number of eligible children dependents
        person = tax_unit.members
        age_eligible = person("age", period) < p.age_limit
        dependent = person("is_tax_unit_dependent", period)
        age_eligible_dependent = age_eligible & dependent
        count_eligible = tax_unit.sum(age_eligible_dependent)

        # Calculate total child tax credit
        return count_eligible * amount_per_qualifying_child



from policyengine_us.model_api import *


class is_father(Variable):
    value_type = bool
    entity = Person
    label = "Is a father"
    definition_period = YEAR

    def formula(person, period, parameters):
        # In the absence of relationship identifiers, check if the person is
        # male and has some children in their own household (provided in the
        # CPS).
        female = person("is_female", period)
        has_children = person("is_parent", period)
        return ~female & has_children



from policyengine_us.model_api import *


class nyc_cdcc_age_restricted_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Childcare expenses for children under NYC CDCC age limit"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # For now define this as total childcare expenses times
        # the share of children in the tax unit under the NYC CDCC age limit.

        # Get the CDCC parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.cdcc

        children = tax_unit("tax_unit_children", period)
        person = tax_unit.members
        qualifying_child = person("age", period) < p.child_age_restriction
        qualifying_children = tax_unit.sum(qualifying_child)
        tax_unit_childcare_expenses = tax_unit(
            "tax_unit_childcare_expenses", period
        )
        # avoid divide-by-zero warnings by not using where() function
        qualifying_child_share = np.zeros_like(children)
        mask = children > 0
        qualifying_child_share[mask] = (
            qualifying_children[mask] / children[mask]
        )
        return tax_unit_childcare_expenses * qualifying_child_share



from policyengine_us.model_api import *


class co_ccap_eligible_children(Variable):
    value_type = int
    entity = SPMUnit
    label = "Number of children eligible for Colorado Child Care Assistance Program"
    definition_period = MONTH
    defined_for = StateCode.CO
    adds = ["co_ccap_child_eligible"]



from policyengine_us.model_api import *


class ca_fera(Variable):
    value_type = float
    entity = Household
    definition_period = YEAR
    unit = USD
    label = "California FERA"
    documentation = "California's FERA program provides this electricity discount to eligible households."
    reference = "https://www.cpuc.ca.gov/industries-and-topics/electrical-energy/electric-costs/care-fera-program"
    defined_for = StateCode.CA

    def formula(household, period, parameters):
        is_eligible = household("ca_fera_eligible", period)
        amount = household("ca_fera_amount_if_eligible", period)
        return is_eligible * amount



from policyengine_us.model_api import *


class is_married(Variable):
    value_type = bool
    entity = Family
    label = "Married"
    documentation = "Whether the adults in this family are married."
    definition_period = YEAR

    def formula(family, period, parameters):
        person = family.members
        return family.any(person("is_tax_unit_spouse", period))



from policyengine_us.model_api import *


class poverty_gap(Variable):
    value_type = float
    entity = SPMUnit
    label = "poverty gap"
    unit = USD
    documentation = "Difference between household income and poverty line."
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        income = spm_unit("spm_unit_net_income", period)
        poverty_threshold = spm_unit("spm_unit_spm_threshold", period)
        return max_(poverty_threshold - income, 0)



from policyengine_us.model_api import *


class az_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        before_non_refundable_credits = tax_unit(
            "az_income_tax_before_non_refundable_credits", period
        )
        non_refundable_credits = tax_unit("az_non_refundable_credits", period)
        return max_(before_non_refundable_credits - non_refundable_credits, 0)



from policyengine_us.model_api import *


class ca_riv_share_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Riverside County Sharing Households Assist Riverside's Energy program (SHARE)"
    definition_period = MONTH
    defined_for = "in_riv"
    reference = "https://riversideca.gov/utilities/residents/assistance-programs/share-english"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.ca.riv.cap.share
        countable_income = spm_unit("ca_riv_share_countable_income", period)
        fpg = spm_unit("spm_unit_fpg", period)
        income_limit = fpg * p.income_limit
        return countable_income <= income_limit



from policyengine_us.model_api import *


class nyc_household_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC Household Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # First get their federal AGI.
        # Technically based on recomputed AGI, which deviates slightly from federal AGI.
        federal_agi = tax_unit("adjusted_gross_income", period)

        # Then get their filing status.
        filing_status = tax_unit("filing_status", period)

        # Then get the household credit part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.household

        # Then get their number of people.
        tax_unit_size = tax_unit("tax_unit_size", period)

        filing_statuses = filing_status.possible_values

        return select(
            [
                filing_status == filing_statuses.SINGLE,
                filing_status == filing_statuses.SEPARATE,
            ],
            [
                # Single filers get a flat amount.
                p.flat_amount.calc(federal_agi, right=True),
                # Separate filers get an amount for each person in the tax
                # unit, varying with AGI.
                p.separate_per_dependent.calc(federal_agi, right=True)
                * tax_unit_size,
            ],
            # Joint, head of household, and surviving spouse filers have a different
            # amount per person, varying with AGI.
            default=p.other_per_dependent.calc(federal_agi, right=True)
            * tax_unit_size,
        )



from policyengine_us.model_api import *


class high_efficiency_electric_home_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "High efficiency electric home rebate"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.doe.high_efficiency_electric_home_rebate
        # Sum (capped) per-expenditure rebates if there are any elements.
        if p.elements == []:
            return 0
        uncapped = add(
            tax_unit, period, ["capped_" + i + "_rebate" for i in p.elements]
        )
        return min_(uncapped, p.cap.total)



from policyengine_us.model_api import *


class is_wic_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    documentation = "Is eligible for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "Is eligible for WIC"
    reference = "https://www.law.cornell.edu/cfr/text/7/246.7#c_1"

    def formula(person, period, parameters):
        meets_income_test = person.spm_unit("meets_wic_income_test", period)
        meets_categorical_test = person(
            "meets_wic_categorical_eligibility", period
        )
        nutritional_risk = person("is_wic_at_nutritional_risk", period)
        # Must be in a valid WIC demographic category per 42 U.S.C. § 1786(d)(1)
        wic_category = person("wic_category", period)
        demographic_eligible = (
            wic_category != wic_category.possible_values.NONE
        )
        return (
            demographic_eligible
            & (meets_income_test | meets_categorical_test)
            & nutritional_risk
        )



from policyengine_us.model_api import *


class small_wind_energy_property_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Qualified small wind energy property expenditures"
    documentation = "Expenditures for property which uses a wind turbine to generate electricity for use in connection with a dwelling unit located in the United States and used as a residence by the taxpayer."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25D#d_4"



from policyengine_us.model_api import *


class oh_lump_sum_retirement_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Ohio lump sum retirement income credit"
    definition_period = YEAR
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=20"
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.credits.retirement.lump_sum
        modified_agi = tax_unit("oh_modified_agi", period)
        distribution_received = (
            tax_unit("form_4972_lumpsum_distributions", period) > 0
        )
        agi_eligible = modified_agi < p.income_limit
        return distribution_received & agi_eligible



from policyengine_us.model_api import *


class wa_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Washington TANF resources eligible"
    definition_period = YEAR
    reference = "https://app.leg.wa.gov/wac/default.aspx?cite=388-470-0005"
    defined_for = StateCode.WA

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.wa.dshs.tanf
        countable_resources = spm_unit("wa_tanf_countable_resources", period)
        return countable_resources <= p.resource_limit



from policyengine_us.model_api import *


class military_disabled_spouse(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = (
        "Tax unit spouse is legally disabled as a result of military service"
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        military_disabled = person(
            "is_fully_disabled_service_connected_veteran", period
        )
        spouse = person("is_tax_unit_spouse", period)
        return tax_unit.any(military_disabled & spouse)



from policyengine_us.model_api import *


class ny_geothermal_energy_system_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York geothermal energy system equipment credit"
    documentation = "The tax credit for a qualified purchase or lease of geothermal energy system equipment, with a 5-year carryover."
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (g)(2)(C)(9)(g-4)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        qualified_expenditures = tax_unit(
            "ny_qualified_geothermal_energy_system_expenditures", period
        )
        p = parameters(
            period
        ).gov.states.ny.tax.income.credits.geothermal_energy_system
        uncapped_credit = qualified_expenditures * p.rate
        return min_(uncapped_credit, p.cap)



from policyengine_us.model_api import *


class in_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana deductions"
    unit = USD
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3"
    defined_for = StateCode.IN

    adds = "gov.states.in.tax.income.deductions.deductions"



from policyengine_us.model_api import *


class il_aabd_need_standard_person(Variable):
    value_type = float
    entity = Person
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) need standard for each person"
    unit = USD
    definition_period = MONTH
    defined_for = "il_aabd_non_financial_eligible_person"
    reference = "https://rockfordha.org/wp-content/uploads/2020/04/Public-Benefits-Quick-Reference-Guide-Updated.pdf#page=1"

    adds = [
        "il_aabd_grant_amount",
        "il_aabd_utility_allowance_person",
        "il_aabd_personal_allowance",
        "il_aabd_shelter_allowance",
    ]



from policyengine_us.model_api import *


class eitc_investment_income_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Meets investment income eligibility for EITC"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/32#i"

    def formula(tax_unit, period, parameters):
        eitc = parameters.gov.irs.credits.eitc(period)
        invinc = tax_unit("eitc_relevant_investment_income", period)
        return invinc <= eitc.phase_out.max_investment_income



from policyengine_us.model_api import *


class pr_eligible_dependent_for_exemption(Variable):
    value_type = bool
    entity = Person
    label = "Puerto Rico eligible dependent for dependent exemption"
    reference = "https://hacienda.pr.gov/sites/default/files/inst_individuals_2023.pdf#page=28"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PR

    def formula(person, period, parameters):
        student_eligible = person("pr_exemptions_is_eligible_student", period)
        non_student_eligible = person(
            "pr_exemptions_is_eligible_nonstudent", period
        )
        return student_eligible | non_student_eligible



from policyengine_us.model_api import *


class tanf_fpg(Variable):
    value_type = float
    entity = SPMUnit
    label = "TANF federal poverty guideline"
    unit = USD
    documentation = (
        "The federal poverty guideline used to determine TANF eligibility."
    )
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/42/9902"

    def formula(spm_unit, period, parameters):
        # We can use tanf_unit_size for more accurate model in the future
        n = spm_unit("spm_unit_size", period.this_year)
        state_group = spm_unit.household("state_group_str", period.this_year)
        year = period.start.year
        month = period.start.month
        if month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p_fpg = parameters(instant_str).gov.hhs.fpg
        p1 = p_fpg.first_person[state_group] / MONTHS_IN_YEAR
        pn = p_fpg.additional_person[state_group] / MONTHS_IN_YEAR
        return p1 + pn * (n - 1)



from policyengine_us.model_api import *


class il_bcc_age_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois Health Benefits for Persons with Breast or Cervical Cancer age eligible"
    definition_period = YEAR
    reference = ("https://www.dhs.state.il.us/page.aspx?item=33528",)
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.hfs.bcc.eligibility
        age = person("age", period)
        return age < p.age



from policyengine_us.model_api import *


class hi_act_115_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii ACT 115 rebate"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = "https://tax.hawaii.gov/act-115-ref/"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.credits.act_115_rebate
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        federal_agi = tax_unit("adjusted_gross_income", period)
        return select(
            [
                filing_status == statuses.JOINT,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.SURVIVING_SPOUSE,
            ],
            [
                p.joint.calc(federal_agi),
                p.single.calc(federal_agi),
                p.single.calc(federal_agi),
                p.joint.calc(federal_agi),
            ],
            default=p.single.calc(federal_agi),
        )



from policyengine_us.model_api import *


class payroll_tax_gross_wages(Variable):
    value_type = float
    entity = Person
    label = "Gross wages and salaries for payroll taxes"
    definition_period = YEAR
    unit = USD
    adds = ["irs_employment_income"]



from policyengine_us.model_api import *


class or_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html"  # 316.695 (1)(d)
    defined_for = StateCode.OR

    adds = ["itemized_deductions_less_salt", "capped_property_taxes"]



from policyengine_us.model_api import *


class wv_sctc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the West Virginia senior citizens tax credit"
    reference = (
        "https://code.wvlegislature.gov/11-21-21/"
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.pdf#page=27 "
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=35"
    )
    definition_period = YEAR
    defined_for = StateCode.WV

    # The senior citizens tax credit is used to calculate the Homestead access property tax credit
    # and provides a credit against property taxes as opposed to income taxes
    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)
        wv_homestead_exemption = tax_unit("wv_homestead_exemption", period)

        p_sctc = parameters(period).gov.states.wv.tax.income.credits.sctc
        p_homestead = parameters(
            period
        ).gov.states.wv.tax.income.exemptions.homestead_exemption

        fpg = tax_unit("tax_unit_fpg", period)
        income_threshold = p_sctc.fpg_percentage * fpg

        meets_agi_condition = federal_agi <= income_threshold
        meets_homestead_exemption_condition = (
            wv_homestead_exemption == p_homestead.max_amount
        )

        return meets_agi_condition & meets_homestead_exemption_condition



from policyengine_us.model_api import *


class ut_military_retirement_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Utah military retirement credit"
    definition_period = YEAR
    reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1043.html"
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        # Has military retirement pay
        military_retirement_pay = add(
            tax_unit, period, ["military_retirement_pay"]
        )
        has_military_retirement_pay = military_retirement_pay > 0
        # Cannot claim if claiming the retirement credit (code 18)
        claims_retirement_credit = tax_unit(
            "ut_claims_retirement_credit", period
        )
        return has_military_retirement_pay & ~claims_retirement_credit



from policyengine_us.model_api import *


class filer_adjusted_earnings(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Filer earned income adjusted for self-employment tax"
    unit = USD

    def formula(tax_unit, period, parameters):
        return tax_unit_non_dep_sum("adjusted_earnings", tax_unit, period)



from policyengine_us.model_api import *


class il_aabd_financial_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) eligible person due to financial criteria"
    definition_period = MONTH
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        asset_eligible = person.spm_unit(
            "il_aabd_asset_value_eligible", period
        )
        total_needs = person("il_aabd_need_standard_person", period)
        countable_income = person("il_aabd_countable_income", period)
        income_eligible = countable_income <= total_needs
        return income_eligible & asset_eligible



from policyengine_us.model_api import *


class pr_agi_person(Variable):
    value_type = float
    entity = Person
    label = "Puerto Rico adjusted gross income person level"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PR
    reference = (
        "https://hacienda.pr.gov/sites/default/files/individuals_2024_rev._jul_12_24_9-30-24_informative.pdf#page=2",
        "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-a/30103/",
    )

    def formula(person, period, parameters):
        filing_status = person.tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        agi_taxunit = person.tax_unit("pr_agi", period)
        frac = where(filing_status == filing_statuses.JOINT, 0.5, 1.0)
        return agi_taxunit * frac



from policyengine_us.model_api import *


class tx_ccs_payment_rate(Variable):
    value_type = float
    entity = Person
    label = "Texas Child Care Services (CCS) payment amount"
    unit = USD
    definition_period = MONTH
    defined_for = "tx_ccs_eligible_child"
    reference = "https://www.twc.texas.gov/sites/default/files/ccel/docs/bcy25-board-max-provider-payment-rates-4-age-groups-twc.pdf#page=9"

    def formula(person, period, parameters):
        # Get the household's workforce board region
        household = person.household
        region = household("tx_ccs_workforce_board_region", period)

        # Get the payment parameters
        p = parameters(period).gov.states.tx.twc.ccs.payment

        # Get the provider characteristics
        provider_type = person("tx_ccs_provider_type", period)
        provider_rating = person("tx_ccs_provider_rating", period)
        age_category = person("tx_ccs_child_age_category", period)
        care_schedule = person("tx_ccs_care_schedule", period)

        # Calculate monthly payment based on region-specific rates
        attending_days_per_month = person(
            "childcare_attending_days_per_month", period.this_year
        )
        return (
            p.rates[region][provider_type][provider_rating][age_category][
                care_schedule
            ]
            * attending_days_per_month
        )



from policyengine_us.model_api import *


class PellGrantHouseholdType(Enum):
    DEPENDENT_SINGLE = "Dependent single"
    DEPENDENT_NOT_SINGLE = "Dependent not single"
    INDEPENDENT_SINGLE = "Independent single"
    INDEPENDENT_NOT_SINGLE = "Independent not single"


class pell_grant_household_type(Variable):
    value_type = Enum
    possible_values = PellGrantHouseholdType
    default_value = PellGrantHouseholdType.INDEPENDENT_SINGLE
    entity = Person
    definition_period = YEAR
    label = "Pell Grant household type"

    def formula(person, period, parameters):
        dependent = person("is_tax_unit_dependent", period)
        joint = person.tax_unit("tax_unit_is_joint", period)

        return select(
            [
                dependent & ~joint,
                dependent & joint,
                ~dependent & ~joint,
                ~dependent & joint,
            ],
            [
                PellGrantHouseholdType.DEPENDENT_SINGLE,
                PellGrantHouseholdType.DEPENDENT_NOT_SINGLE,
                PellGrantHouseholdType.INDEPENDENT_SINGLE,
                PellGrantHouseholdType.INDEPENDENT_NOT_SINGLE,
            ],
        )



from policyengine_us.model_api import *


class pr_low_income_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible unit for the Puerto Rico low income credit"
    definition_period = YEAR
    reference = "https://casetext.com/statute/laws-of-puerto-rico/title-thirteen-taxation-and-finance/subtitle-17-internal-revenue-code-of-2011/part-ii-income-taxes/chapter-1007-credits-against-tax/subchapter-b-refundable-credits/30212-credit-for-low-income-individuals-older-than-sixty-five-65-years-of-age"
    defined_for = "pr_low_income_credit_eligible_people"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.territories.pr.tax.income.credits.low_income
        eligible_people = tax_unit(
            "pr_low_income_credit_eligible_people", period
        )
        income = tax_unit("pr_gross_income_person", period)
        income_limit = p.income_limit.calc(eligible_people)
        return income <= income_limit



from policyengine_us.model_api import *


class ny_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY income deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/613"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        return max_(
            tax_unit("ny_itemized_deductions", period),
            tax_unit("ny_standard_deduction", period),
        )



from policyengine_us.model_api import *
from policyengine_us.variables.household.income.person.fsla_overtime_occupation_exemption_category import (
    OvertimeExemptionCategory,
)


class fsla_overtime_salary_threshold(Variable):
    value_type = float
    entity = Person
    label = "FSLA applicable salary threshold to determine eligibility for overtime pay"
    reference = "https://www.law.cornell.edu/cfr/text/29/541.600"
    definition_period = YEAR
    unit = USD

    def formula_2014(person, period, parameters):
        p = parameters(period).gov.irs.income.exemption.overtime
        category = person(
            "fsla_overtime_occupation_exemption_category", period
        )

        # Initialize all thresholds to HCE - this applies to everyone by default
        # except for those who are always exempt (military, never worked)
        threshold = np.full_like(category, p.hce_salary_threshold, dtype=float)

        # Special category thresholds (these override HCE if they're lower)
        is_computer = category == OvertimeExemptionCategory.COMPUTER_SCIENTIST
        is_exec_admin = (
            category == OvertimeExemptionCategory.EXECUTIVE_ADMINISTRATIVE
        )
        is_farmer_fisher = category == OvertimeExemptionCategory.FARMER_FISHER

        threshold = np.where(
            is_computer,
            np.minimum(
                p.computer_salary_threshold
                * WEEKS_IN_YEAR
                * p.hours_threshold,
                p.hce_salary_threshold,
            ),
            threshold,
        )
        threshold = np.where(
            is_exec_admin | is_farmer_fisher,
            np.minimum(
                p.salary_basis_threshold * WEEKS_IN_YEAR,
                p.hce_salary_threshold,
            ),
            threshold,
        )

        # Always exempt categories have threshold of 0 (any income means exempt)
        is_always_exempt = (category == OvertimeExemptionCategory.MILITARY) | (
            category == OvertimeExemptionCategory.NEVER_WORKED
        )
        threshold = np.where(is_always_exempt, 0, threshold)

        return threshold



from policyengine_us.model_api import *


class older_spouse_birth_year(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    label = "Birth year of head or spouse of tax unit depending on which is greater"
    documentation = "Birth year of taxpayer (i.e. primary adult) or spouse (i.e. secondary adult if present), depending on which is greater. "
    unit = "year"

    def formula(tax_unit, period, parameters):
        return period.start.year - tax_unit("greater_age_head_spouse", period)



from policyengine_us.model_api import *


class relative_income_change(Variable):
    value_type = float
    entity = Person
    label = "relative income change"
    unit = "/1"
    definition_period = YEAR
    requires_computation_after = "labor_supply_behavioral_response"

    def formula(person, period, parameters):
        simulation = person.simulation
        measurement_branch = simulation.get_branch("lsr_measurement")
        baseline_branch = simulation.get_branch("baseline").get_branch(
            "baseline_lsr_measurement"
        )
        baseline_person = baseline_branch.populations["person"]
        baseline_net_income = baseline_person.household(
            "household_net_income", period
        )
        measurement_person = measurement_branch.populations["person"]
        net_income = measurement_person.household(
            "household_net_income", period
        )
        p = parameters(period).gov.simulation.labor_supply_responses
        # _c suffix for "clipped"
        baseline_net_income_c = np.clip(baseline_net_income, 1, None)
        net_income_c = np.clip(net_income, 1, None)
        relative_change = (
            net_income_c - baseline_net_income_c
        ) / baseline_net_income_c
        return np.clip(
            relative_change, -p.bounds.income_change, p.bounds.income_change
        )



import collections
import requests
from pathlib import Path
import pandas as pd
import subprocess
import stat
from io import StringIO
from policyengine_core.data import Dataset
from policyengine_core.taxbenefitsystems import TaxBenefitSystem
from policyengine_us import Microsimulation
from tqdm import tqdm
from argparse import ArgumentParser
import platform

PLATFORM_MAPPING = {
    "Windows": "windows",
    "Linux": "unix",
    "Darwin": "osx",
}

taxsim_platform = PLATFORM_MAPPING[platform.system()]


class TaxSim35:
    """TAXSIM 35 Internet version: http://taxsim.nber.org/taxsim35/"""

    EXECUTABLE_URL = f"https://taxsim.nber.org/stata/taxsim35/taxsim35-{taxsim_platform}.exe"
    folder = Path(__file__).parent.absolute()
    executable_path = folder / "taxsim35.exe"
    INPUT_VARIABLES = [
        "taxsimid",
        "year",
        "state",
        "mstat",
        "page",
        "sage",
        "depx",
        # "age1",
        # "age2",
        # "age3",
        "dep13",
        "dep17",
        "dep18",
        "pwages",
        "swages",
        "psemp",
        "ssemp",
        "dividends",
        "intrec",
        "stcg",
        "ltcg",
        # "otherprop",
        # "nonprop",
        "pensions",
        "gssi",
        "pui",
        "ui",
        # "transfers",
        # "rentpaid",
        # "proptax",
        # "otheritem",
        "childcare",
        # "mortgage",
        "scorp",
        "pbusinc",
        "pprofinc",
        "sbusinc",
        "sprofinc",
    ]
    UNIMPLEMENTED_VARIABLES = [
        "otherprop",
        "nonprop",
        "transfers",
        "rentpaid",
        "proptax",
        "otheritem",
        "mortgage",
    ]
    OUTPUT_VARIABLES = [
        "fiitax",
        "siitax",
    ]
    policyengine_us_INPUT_VARIABLES = [
        "mars",
        "employment_income",
        "age",
        "is_tax_unit_head",
        "is_tax_unit_spouse",
        "is_tax_unit_dependent",
        "self_employment_income",
        "dividend_income",
        "pension_income",
        "social_security",
        "unemployment_compensation",
        "tax_unit_childcare_expenses",
        "short_term_capital_gains",
        "long_term_capital_gains",
        "taxable_interest_income",
        "partnership_s_corp_income",
        "tax_unit_id",
        "person_tax_unit_id",
    ]
    NEUTRALIZED_VARIABLES = [
        "snap",
        "ssi",
        "tanf",
        "wic",
    ]

    def __init__(self):
        self.ensure_executable_is_ready()

    @property
    def executable_is_stored(self):
        return (self.folder / "taxsim35.exe").exists()

    def ensure_executable_is_ready(self):
        if not self.executable_is_stored:
            # Download the .exe file and save to the folder
            response = requests.get(self.EXECUTABLE_URL)
            with open(self.folder / "taxsim35.exe", "wb") as f:
                f.write(response.content)
            self.executable_path.chmod(
                self.executable_path.stat().st_mode | stat.S_IEXEC
            )

    def calculate(self, input_data: dict) -> dict:
        if isinstance(input_data, pd.DataFrame):
            input_csv = input_data
        else:
            input_csv = pd.DataFrame(
                {
                    col: (
                        [value]
                        if not isinstance(value, collections.Sequence)
                        else value
                    )
                    for col, value in input_data.items()
                }
            )
        input_csv_text = input_csv.to_csv(index=False)
        # We would usually run: taxsim35.exe < input.csv > output.csv.
        # But we're in a Python interface, so we just want the output as a DataFrame.

        process = subprocess.Popen(
            [str(self.executable_path.absolute())],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        result = process.communicate(input_csv_text.encode())[0].decode()
        try:
            return pd.read_csv(StringIO(result))
        except:
            # Full text results (idtl = 5)
            outputs = result.split("Federal Tax Calculation:")[1:]
            column_names = [f"v{i}" for i in range(10, 30)]
            column_names = [i for i in column_names if i != "v23"]
            output_data = {column: [] for column in column_names}
            for output in tqdm(
                outputs, desc="Parsing detailed outputs from TAXSIM"
            ):
                columns_used = []
                for line in output.split("\n"):
                    if any(
                        [
                            name.replace("v", "") + "." in line
                            for name in column_names
                        ]
                    ):
                        number = line.split(".")[0].strip()
                        value = line.split(".")[1].split("  ")[-1]
                        name = f"v{number}"
                        if name in output_data and name not in columns_used:
                            output_data[name].append(int(value))
                            columns_used.append(name)
            return pd.DataFrame(output_data)

    def generate_from_microsimulation(
        self,
        dataset: Dataset,
        year: int,
        number: int,
        return_dataframe: bool = False,
        drop_zeros: bool = False,
    ):
        sim = Microsimulation(dataset=dataset, dataset_year=2021)

        # Neutralise benefit variables, since these are not simulated in TAXSIM
        for variable in self.NEUTRALIZED_VARIABLES:
            sim.tax_benefit_system.neutralize_variable(variable)

        system: TaxBenefitSystem = sim.tax_benefit_system
        openfisca_named_taxsim_input_variables = [
            f"taxsim_{variable}" for variable in self.INPUT_VARIABLES
        ]
        openfisca_named_taxsim_output_variables = [
            f"taxsim_{variable}" for variable in self.OUTPUT_VARIABLES
        ]
        full_input_df = pd.DataFrame(
            sim.df(openfisca_named_taxsim_input_variables, period=year)
        )
        if number is not None:
            full_input_df = full_input_df.sample(n=number * 100)
        input_df = full_input_df.reset_index(drop=True)
        input_df.columns = self.INPUT_VARIABLES
        input_df["idtl"] = 2  # Set full output
        for variable in self.UNIMPLEMENTED_VARIABLES:
            input_df[variable] = 0
        taxsim_df = self.calculate(input_df).reset_index(drop=True)
        taxsim_df = taxsim_df.rename(
            columns={col: f"taxsim_{col}" for col in taxsim_df.columns}
        )
        taxsim_df = pd.concat([input_df, taxsim_df], axis=1)
        variables = system.variables
        i = 0
        test_str = ""
        tax_unit_number = 1
        # Shuffle the dataframe
        is_non_zero = (
            taxsim_df[openfisca_named_taxsim_output_variables].sum(axis=1) > 0
        )
        if drop_zeros:
            taxsim_df = taxsim_df[is_non_zero]
        taxsim_df = taxsim_df.sample(frac=1).reset_index(drop=True)
        if return_dataframe:
            return taxsim_df
        for tax_unit_id in tqdm(
            taxsim_df.taxsim_taxsimid, desc="Writing YAML tests"
        ):
            if number is not None and i >= number:
                break
            i += 1
            test_str += f"- name: Tax unit {tax_unit_number:,.0f} (CPS ID {tax_unit_id}) matches TAXSIM35 outputs\n  absolute_error_margin: 1\n  period: {year}\n  input:\n    people:\n"
            tax_unit_number += 1
            person_id = sim.calc("person_id").values
            person_tax_unit_id = sim.calc(
                "tax_unit_id", map_to="person"
            ).values
            tax_unit_ids = sim.calc("tax_unit_id").values
            people_in_tax_unit = person_id[person_tax_unit_id == tax_unit_id]
            person_number = 1
            for person in people_in_tax_unit:
                test_str += f"      person_{person_number}:\n"
                person_number += 1
                for variable_name in (
                    self.policyengine_us_INPUT_VARIABLES
                    + openfisca_named_taxsim_input_variables
                ):
                    if variables[variable_name].entity.key == "person":
                        value = sim.calc(
                            variable_name, map_to="person"
                        ).values[person_id == person][0]
                        try:
                            test_str += (
                                f"        {variable_name}: {value:_.0f}\n"
                            )
                        except:
                            test_str += f"        {variable_name}: {value}\n"
            test_str += f"    tax_units:\n      tax_unit:\n        members: [{','.join(['person_' + str(p) for p in range(1, person_number)])}]\n"
            for variable_name in (
                self.policyengine_us_INPUT_VARIABLES
                + openfisca_named_taxsim_input_variables
            ):
                if variables[variable_name].entity.key == "tax_unit":
                    value = sim.calc(variable_name).values[
                        tax_unit_ids == tax_unit_id
                    ][0]
                    try:
                        test_str += f"        {variable_name}: {value:_.0f}\n"
                    except:
                        test_str += f"        {variable_name}: {value}\n"
            test_str += f"  output:\n"
            for variable_name in openfisca_named_taxsim_output_variables:
                if variables[variable_name].entity.key == "tax_unit":
                    value = taxsim_df[variable_name][
                        taxsim_df.taxsim_taxsimid == tax_unit_id
                    ].values[0]
                    try:
                        test_str += f"    {variable_name}: {value:_.0f}\n"
                    except:
                        test_str += f"    {variable_name}: {value}\n"
            test_str += "\n\n"
        return test_str


DEFAULT_DATASET = "hf://policyengine/policyengine-us-data/enhanced_cps_2024.h5"


if __name__ == "__main__":
    parser = ArgumentParser(description="Generate tests from TAXSIM35")
    parser.add_argument(
        "outputs", nargs="*", help="Outputs to generate tests for."
    )
    parser.add_argument(
        "-num", type=int, default=32, help="Number of tests to generate."
    )
    parser.add_argument(
        "-year", type=int, default=2022, help="Year to generate tests for."
    )
    args = parser.parse_args()
    taxsim = TaxSim35()
    taxsim.OUTPUT_VARIABLES = args.outputs
    result = taxsim.generate_from_microsimulation(
        DEFAULT_DATASET, args.year, number=args.num
    )
    print(result)



from policyengine_us.model_api import *


class ny_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY income tax before credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY

    adds = ["ny_main_income_tax", "ny_supplemental_tax"]



from policyengine_us.model_api import *


class employee_social_security_tax(Variable):
    value_type = float
    entity = Person
    label = "employee-side OASDI payroll tax"
    documentation = "Total liability for employee-side OASDI payroll tax."
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.payroll.social_security.rate
        return p.employee * person(
            "taxable_earnings_for_social_security", period
        )



from policyengine_us.model_api import *


class ia_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ia.tax.income.married_filing_separately_on_same_return
        if p.availability:
            return where(
                tax_unit("ia_files_separately", period),
                tax_unit("ia_income_tax_indiv", period),
                tax_unit("ia_income_tax_joint", period),
            )
        else:
            return tax_unit("ia_income_tax_consolidated", period)



from policyengine_us.model_api import *


class tx_tanf_age_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Age-eligible child for Texas TANF based on demographics"
    definition_period = MONTH
    reference = (
        "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-220-tanf",
        "https://www.law.cornell.edu/regulations/texas/1-TAC-372-307",
    )
    defined_for = StateCode.TX

    def formula(person, period, parameters):
        p = parameters(period).gov.states.tx.tanf.age_threshold
        age = person("monthly_age", period)
        is_dependent = person("is_tax_unit_dependent", period)
        is_full_time_student = person("is_full_time_student", period)

        age_eligible = where(
            is_full_time_student,
            age < p.student_dependent,
            age < p.minor_child,
        )

        return age_eligible & is_dependent



from policyengine_us.model_api import *


class is_in_k12_nonpublic_school(Variable):
    value_type = bool
    entity = Person
    label = "Is in a K-12 nonpublic school fulltime"
    definition_period = YEAR



from policyengine_us.model_api import *


class slcsp_age_curve_amount_person(Variable):
    value_type = float
    entity = Person
    label = (
        "Second-lowest ACA silver-plan cost, for people in age curve states"
    )
    unit = USD
    definition_period = MONTH
    defined_for = "is_aca_ptc_eligible"

    def formula(person, period, parameters):
        state_code = person.household("state_code_str", period)
        age = person("monthly_age", period)
        base_cost = person.household("slcsp_age_0", period)

        p = parameters(period).gov.aca.age_curves

        # Handle other states with regular bracket structures
        multiplier = select(
            [
                state_code == "AL",
                state_code == "DC",
                state_code == "MA",
                state_code == "MN",
                state_code == "MS",
                state_code == "OR",
                state_code == "UT",
            ],
            [
                p.al.calc(age),
                p.dc.calc(age),
                p.ma.calc(age),
                p.mn.calc(age),
                p.ms.calc(age),
                p["or"].calc(age),
                p.ut.calc(age),
            ],
            default=p.default.calc(age),
        )
        age_curve_applies = person.tax_unit("slcsp_age_curve_applies", period)
        return base_cost * multiplier * age_curve_applies



from policyengine_us.model_api import *


class additional_medicare_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Additional Medicare Tax"
    unit = USD
    documentation = (
        "Additional Medicare Tax from Form 8959 (included in payrolltax)"
    )

    def formula(tax_unit, period, parameters):
        amc = parameters(period).gov.irs.payroll.medicare.additional
        # Wage and self-employment income are taxed the same.
        ELEMENTS = ["irs_employment_income", "taxable_self_employment_income"]
        wages_plus_se = add(tax_unit, period, ELEMENTS)
        exclusion = amc.exclusion[tax_unit("filing_status", period)]
        base = max_(0, wages_plus_se - exclusion)
        return amc.rate * base



from policyengine_us.model_api import *


class taxable_401k_distributions(Variable):
    value_type = float
    entity = Person
    label = "Taxable 401(k) distributions"
    unit = USD
    documentation = (
        "Taxable distributions from 401k accounts (typically traditional)."
    )
    definition_period = YEAR



from policyengine_us.model_api import *


class meets_snap_asset_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets SNAP asset test"
    documentation = "Whether the SPM unit's financial resources are within SNAP's allowable limit"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        has_elderly_or_disabled = spm_unit("has_usda_elderly_disabled", period)
        asset_test = parameters(period).gov.usda.snap.asset_test
        assets = spm_unit("snap_assets", period)
        asset_limit = where(
            has_elderly_or_disabled,
            asset_test.limit.elderly_disabled,
            asset_test.limit.standard,
        )
        return assets <= asset_limit



from policyengine_us.model_api import *


class snap_self_employment_income_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "All self-employment income expenses for the SNAP self-employment deduction"
    unit = USD
    definition_period = YEAR


# Each state defines deductible self-employment income expenses for SNAP.
# E.g. see Mississippi: https://www.law.cornell.edu/regulations/mississippi/18-Miss-Code-R-SS-14-20-4
# The relevant expenses are often linked to TANF self-employment income expenses.



from policyengine_us.model_api import *


class in_aged_low_agi_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana aged and low-AGI exemptions"
    unit = USD
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"  # (a)(5)(C)
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["in"].tax.income.exemptions
        filing_status = tax_unit("filing_status", period)
        threshold = p.aged_low_agi.threshold[filing_status]
        aged_low_agi_exemption = p.aged_low_agi.amount
        federal_agi = tax_unit("adjusted_gross_income", period)
        aged_head = tax_unit("aged_head", period).astype(int)
        aged_spouse = tax_unit("aged_spouse", period).astype(int)
        income_eligible = federal_agi < threshold
        return (
            income_eligible
            * (aged_head + aged_spouse)
            * aged_low_agi_exemption
        )



from policyengine_us.model_api import *


class vt_child_care_contributions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont child care contributions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VT

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.vt.tax.income.child_care_contributions
        if p.applies:
            income = add(tax_unit, period, ["self_employment_income"])
            applicable_income = max_(0, income) * p.rate.income
            return applicable_income * p.rate.contributions
        return 0



from policyengine_us.model_api import *


def create_abolish_snap_net_income_test() -> Reform:
    class meets_snap_net_income_test(Variable):
        value_type = bool
        entity = SPMUnit
        label = "Meets SNAP net income test"
        documentation = "Whether this SPM unit meets the SNAP net income test"
        definition_period = MONTH
        reference = (
            "https://www.law.cornell.edu/uscode/text/7/2017#a",
            "https://www.law.cornell.edu/uscode/text/7/2014#c",
        )

        def formula(spm_unit, period, parameters):
            return True

    class reform(Reform):
        def apply(self):
            self.update_variable(meets_snap_net_income_test)

    return reform


def create_abolish_snap_net_income_test_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_abolish_snap_net_income_test()

    p = parameters(period).gov.contrib.snap.abolish_net_income_test

    if p.in_effect:
        return create_abolish_snap_net_income_test()
    else:
        return None


abolish_snap_net_income_test = create_abolish_snap_net_income_test_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class il_smib_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois SMIB Buy-In Program eligible"
    definition_period = MONTH
    documentation = (
        "Eligible for Illinois Supplementary Medical Insurance Benefit "
        "(SMIB) Buy-In Program. The program pays Medicare Part B premiums "
        "for eligible individuals receiving AABD, TANF, SSI, or who qualify "
        "as QMB/SLMB/QI under the Medicare Savings Program."
    )
    reference = (
        "https://www.ilga.gov/commission/jcar/admincode/089/089001200D00700R.html",
        "https://www.dhs.state.il.us/page.aspx?item=18685",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        # Must be Medicare eligible (or meet other SMIB criteria)
        is_medicare = person("is_medicare_eligible", period.this_year)
        # Check categorical eligibility (AABD, TANF, SSI)
        categorical_eligible = person("il_smib_categorical_eligible", period)
        # Note: QMB/SLMB/QI eligibility will come from MSP when that PR merges
        # For now, categorical eligibility is the primary path
        return is_medicare & categorical_eligible



from policyengine_us.model_api import *


class co_tanf_need_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado TANF need standard"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        # Get number of children and adults in SPM unit.
        children = spm_unit("co_tanf_count_children", period)
        adults = spm_unit("spm_unit_count_adults", period)
        # Cap them at the maximum specified in the tables.
        capped_children = min_(children, 10).astype(int)
        capped_adults = min_(adults, 2).astype(int)
        # Calculate additional children beyond the maximum in tables.
        additional_children = children - capped_children
        # Get the relevant part of the parameter tree.
        p = parameters(period).gov.states.co.cdhs.tanf.need_standard
        # Look up the main need standard for the number of adults and (capped)
        # children.
        base = p.main[capped_adults][capped_children]
        # Add the additional need standard for the additional children.
        additional_need_standard = p.additional_child * additional_children
        monthly = base + additional_need_standard
        # Return annual value.
        return monthly * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class ok_federal_eitc_investment_income_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Meets investment income eligibility for EITC"
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/oklahoma/title-68/section-68-2357-43/"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        eitc = parameters.gov.irs.credits.eitc(f"2020-01-01")
        invinc = tax_unit("eitc_relevant_investment_income", period)
        return invinc <= eitc.phase_out.max_investment_income



from policyengine_us.model_api import *


class ctc_maximum(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum CTC"
    unit = USD
    documentation = "Maximum value of the Child Tax Credit, before phase-out."
    definition_period = YEAR

    adds = ["ctc_individual_maximum"]



from policyengine_us.model_api import *


class ms_files_separately(Variable):
    value_type = bool
    entity = TaxUnit
    label = "married couple files separately on Mississippi tax return"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MS

    def formula(tax_unit, period, parameters):
        itax_indiv = add(
            tax_unit, period, ["ms_income_tax_before_credits_indiv"]
        )
        itax_joint = add(
            tax_unit, period, ["ms_income_tax_before_credits_joint"]
        )
        return itax_indiv < itax_joint



from policyengine_us.model_api import *


class unadjusted_basis_qualified_property(Variable):
    value_type = float
    entity = Person
    label = "Unadjusted basis for qualified property"
    unit = USD
    documentation = "Share of unadjusted basis upon acquisition of all property held by qualified pass-through businesses."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#b_6"



from policyengine_us.model_api import *


class in_use_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana use tax"
    unit = USD
    definition_period = YEAR
    reference = "https://iga.in.gov/laws/2021/ic/titles/6#6-2.5-3"
    defined_for = StateCode.IN



from policyengine_us.model_api import *


class la_agi_exempt_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana income that is exempt from the adjusted gross income"
    defined_for = StateCode.LA
    unit = USD
    reference = (
        "https://revenue.louisiana.gov/TaxForms/IT540i(2021)%20Instructions.pdf#page=9",
        "https://revenue.louisiana.gov/TaxForms/IT540WEB(2021)%20F.pdf",
    )
    definition_period = YEAR

    # Functions as subtractions.
    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.la.tax.income.exempt_income
        total_exempt_income = add(tax_unit, period, p.sources)
        if p.reduction.in_effect:
            # Option 1 is to reduce the exempt income by a marginal rate
            reduced_exempt_income = p.reduction.rate.calc(total_exempt_income)
            # Option 2 is to reduce the federal tax deduction by the percentage of
            # exempt income to AGI
            agi = tax_unit("adjusted_gross_income", period)
            exempt_income_rate = np.zeros_like(agi)
            mask = agi != 0
            exempt_income_rate[mask] = total_exempt_income[mask] / agi[mask]
            # The step by step calculation is as follows:
            # Line 1, representing federal AGI is divided by Line 4H
            # which is the total exempt income. This rate (exempt_income_rate) is multiplied by
            # Line 9 of the Form IT-540, representing the federal tax deduction
            federal_tax_deduction = tax_unit(
                "la_federal_tax_deduction", period
            )
            # Multiply the federal tax deduction by the exempt income rate
            reduced_federal_tax_deduction = (
                federal_tax_deduction * exempt_income_rate
            )
            # The smaller of the two options is applied if taxable income exceeds zero
            # Mentioned from the IT-540 instructions, Page 9, Line 4I
            smaller_adjustment = min_(
                reduced_federal_tax_deduction, reduced_exempt_income
            )
            # The smaller of the two options is applied if taxable income exceeds zero
            federal_tax_deduction_eligible = federal_tax_deduction > 0
            final_exempt_income_reduction = where(
                federal_tax_deduction_eligible,
                smaller_adjustment,
                reduced_exempt_income,
            )
            return max_(total_exempt_income - final_exempt_income_reduction, 0)
        return total_exempt_income



from policyengine_us.model_api import *


class la_general_relief_rent_contribution(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Los Angeles County General Relief rent contribution"
    definition_period = MONTH
    defined_for = "la_general_relief_eligible"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"

    def formula(spm_unit, period, parameters):
        married = add(spm_unit, period, ["is_married"])
        p = parameters(period).gov.local.ca.la.general_relief
        # If filers are receiving the housing subsidy
        # they are obligate to commit an amount of their GR towards rent
        # which is deducted here and added to the housing subsidy amount
        receive_housing_subsidy = (
            spm_unit("la_general_relief_housing_subsidy", period) > 0
        )
        rent_contributions = where(
            married,
            p.housing_subsidy.rent_contribution.married,
            p.housing_subsidy.rent_contribution.single,
        )

        return where(receive_housing_subsidy, rent_contributions, 0)



from policyengine_us.model_api import *


class ArizonaFilingStatus(Enum):
    SINGLE = "Single"
    SEPARATE = "Separate"
    HEAD_OF_HOUSEHOLD = "Head of household"
    JOINT = "Joint"


class az_filing_status(Variable):
    value_type = Enum
    entity = TaxUnit
    possible_values = ArizonaFilingStatus
    default_value = ArizonaFilingStatus.SINGLE
    definition_period = YEAR
    documentation = (
        "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01001.htm"  # Section 43-1001(5)
        "https://azdor.gov/forms/individual/form-140a-arizona-resident-personal-income-tax-booklet"  # Page 5 (page 9 of pdf), Box 5.
    )

    label = "Arizona filing status"

    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        us_filing_status = tax_unit("filing_status", period)
        fsvals = us_filing_status.possible_values
        return select(
            [
                us_filing_status == fsvals.JOINT,
                us_filing_status == fsvals.SINGLE,
                us_filing_status == fsvals.SEPARATE,
            ],
            [
                # In Arizona, surviving spouse filers are treated as heads of households.
                ArizonaFilingStatus.JOINT,
                ArizonaFilingStatus.SINGLE,
                ArizonaFilingStatus.SEPARATE,
            ],
            default=ArizonaFilingStatus.HEAD_OF_HOUSEHOLD,
        )



from policyengine_us.model_api import *


class wv_homestead_excess_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = (
        "Eligible for the West Virginia homestead excess property tax credit"
    )
    reference = (
        "https://code.wvlegislature.gov/11-21-23/"
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.pdf#page=13"
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=14"
    )
    definition_period = YEAR
    defined_for = StateCode.WV

    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)
        wv_sctc = tax_unit("wv_sctc", period)
        property_tax = add(tax_unit, period, ["real_estate_taxes"])
        wv_ghi = tax_unit("wv_gross_household_income", period)
        wv_tax_unit_fpg = tax_unit("tax_unit_fpg", period)
        p = parameters(period).gov.states.wv.tax.income.credits.heptc.rate
        low_income_guidelines = p.fpg * wv_tax_unit_fpg
        lig_eligible = federal_agi <= low_income_guidelines
        property_tax_value = property_tax - wv_sctc
        ghi_amount = p.household_income * wv_ghi
        property_tax_eligible = property_tax_value > ghi_amount
        return lig_eligible & property_tax_eligible



from policyengine_us.model_api import *


class ma_state_supplement(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts State Supplement payment amount"
    definition_period = YEAR
    defined_for = StateCode.MA
    exhaustive_parameter_dependencies = "gov.states.ma.dta.ssp"
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-327-330"
    )

    def formula(person, period, parameters):
        uncapped_ssi = person("uncapped_ssi", period)
        reduction_after_ssi = max_(0, -uncapped_ssi)
        maximum_ss = person("ma_maximum_state_supplement", period)
        state_supplement = max_(0, maximum_ss - reduction_after_ssi)
        abd = person("is_ssi_aged_blind_disabled", period)
        meets_resource_test = person("meets_ssi_resource_test", period)
        eligible = abd & meets_resource_test
        joint_claim = person("ssi_claim_is_joint", period)
        return where(
            joint_claim,
            person.marital_unit.sum(state_supplement) / 2,
            state_supplement * eligible,
        )



from policyengine_us.model_api import *


class me_child_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine child care credit"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://www.mainelegislature.org/legis/statutes/36/title36sec5218.html",
        "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_a_ff.pdf#page=2",
    ]
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.me.tax.income.credits.child_care

        # Get share of expenses that went to step 4 programs
        step_4_share_of_expenses = tax_unit(
            "me_step_4_share_of_child_care_expenses", period
        )
        # Line 2: Divide Federal CDCC according to share of regular vs. Step 4 expenses
        # Maine matches the federal credit taken
        cdcc = tax_unit("cdcc", period)
        # Line 2a: Column A
        cdcc_regular_portion = cdcc * (1 - step_4_share_of_expenses)
        # Line 2a, Column B
        cdcc_step_4_portion = cdcc * step_4_share_of_expenses
        # Line 3, Column A
        regular_child_care_credit = (
            p.share_of_federal_credit.non_step_4 * cdcc_regular_portion
        )
        # Line 3, Column B
        step_4_child_care_credit = (
            p.share_of_federal_credit.step_4 * cdcc_step_4_portion
        )
        return regular_child_care_credit + step_4_child_care_credit



from policyengine_us.model_api import *


class CaCalworksChildCareProviderCategory(Enum):
    CHILD_CARE_CENTER = "Child care center"
    FAMILY_CHILD_CARE_HOME = "Family and child care home"
    LICENSE_EXEMPT = "License exempt"


class ca_calworks_child_care_provider_category(Variable):
    value_type = Enum
    possible_values = CaCalworksChildCareProviderCategory
    default_value = CaCalworksChildCareProviderCategory.CHILD_CARE_CENTER
    entity = Person
    label = "California CalWORKs Child Care provider categroy"
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_8_Regional_Market_Rate_Ceilings%2F1210_8_Regional_Market_Rate_Ceilings.htm%23Contactbc-13&rhtocid=_3_3_8_12"



from policyengine_us.model_api import *


class ma_ccfa_reimbursement_ratio(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts Child Care Financial Assistance (CCFA) reimbursement ratio"
    reference = "https://www.mass.gov/doc/fiscal-year-2025-child-care-financial-assistance-daily-reimbursement-rates/download"
    definition_period = MONTH
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ma.eec.ccfa.reimbursement_rates
        childcare_hours_per_day = person(
            "childcare_hours_per_day", period.this_year
        )
        return p.amount_ratio.calc(childcare_hours_per_day)



from policyengine_us.model_api import *


class snap_dependent_care_deduction(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP dependent care deduction"
    unit = USD
    documentation = "Deduction from SNAP gross income for dependent care"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e_3"

    # Use pre-subsidy expenses to avoid circular dependency
    # (childcare subsidies depend on income, which would create a
    # circular dependency if SNAP deductions depended on subsidized
    # childcare expenses).
    adds = ["pre_subsidy_childcare_expenses"]



from policyengine_us.model_api import *


class dc_power_has_disqualifying_benefits(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Is receiving disqualifying benefits under DC Program on Work, Employment, and Responsibility (POWER)"
    definition_period = YEAR
    defined_for = StateCode.DC
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.72#(d)"
    )

    adds = "gov.states.dc.dhs.power.disqualifying_benefits"



from policyengine_us.model_api import *


class slcsp_rating_area(Variable):
    value_type = int
    entity = Household
    label = "Second-lowest ACA silver-plan cost rating area"
    definition_period = YEAR

    def formula(household, period, parameters):
        return where(
            household("in_la", period),
            household("slcsp_rating_area_la_county", period),
            household("slcsp_rating_area_default", period),
        )
